##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module parses the batch section of the yaml specification.

Currently only the batch worker launch for slurm, lsf or flux
are implemented.
"""
import logging
import os
import subprocess
from typing import Dict, Union

from merlin.spec.specification import MerlinSpec
from merlin.utils import convert_timestring, get_flux_alloc, get_flux_version, get_yaml_var


LOG = logging.getLogger(__name__)


def batch_check_parallel(spec: MerlinSpec) -> bool:
    """
    Check for a parallel batch section in the provided MerlinSpec object.

    This function examines the 'batch' section of the given specification to determine
    whether it is configured for parallel execution. It checks the 'type' attribute
    within the batch section, defaulting to 'local' if not specified. If the type
    is anything other than 'local', the function will return True, indicating that
    parallel processing is enabled.

    Args:
        spec (spec.specification.MerlinSpec): An instance of the
            [`MerlinSpec`][spec.specification.MerlinSpec] class that contains the
            configuration details, including the batch section.

    Returns:
        Returns True if the batch type is set to a value other than 'local',
            indicating that parallel processing is enabled; otherwise, returns False.

    Raises:
        AttributeError: If the 'batch' section is not present in the specification,
            an error is logged and an AttributeError is raised.
    """
    parallel = False

    try:
        batch = spec.batch
    except AttributeError as exc:
        LOG.error("The batch section is required in the specification file.")
        raise exc

    btype = get_yaml_var(batch, "type", "local")
    if btype != "local":
        parallel = True

    return parallel


def check_for_scheduler(scheduler: str, scheduler_legend: Dict[str, str]) -> bool:
    """
    Check which scheduler (Flux, Slurm, LSF, or PBS) is the main scheduler for the cluster.

    This function verifies if the specified scheduler is the main scheduler by executing
    a command associated with it from the provided scheduler legend. It returns a boolean
    indicating whether the specified scheduler is active.

    Args:
        scheduler: A string representing the scheduler to check for. Options include 'flux',
            'slurm', 'lsf', or 'pbs'.
        scheduler_legend: A dictionary containing information related to each scheduler,
            including the command to check its status and the expected output. See
            [`construct_scheduler_legend`][study.batch.construct_scheduler_legend]
            for more information on all the settings this dict contains.

    Returns:
        Returns True if the specified scheduler is the main scheduler for the
            cluster, otherwise returns False.

    Raises:
        FileNotFoundError: If the command associated with the scheduler cannot be found.
        PermissionError: If there are insufficient permissions to execute the command.
    """
    # Check for invalid scheduler
    if scheduler not in ("flux", "slurm", "lsf", "pbs"):
        LOG.warning(f"Invalid scheduler {scheduler} given to check_for_scheduler.")
        return False

    # Try to run the check command provided via the scheduler legend
    try:
        process = subprocess.Popen(  # pylint: disable=R1732
            scheduler_legend[scheduler]["check cmd"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # If the desired output exists, return True. Otherwise, return False
        result = process.stdout.readlines()
        if result and len(result) > 0 and scheduler_legend[scheduler]["expected check output"] in result[0]:
            return True
        return False
    except (FileNotFoundError, PermissionError):
        return False


def get_batch_type(scheduler_legend: Dict[str, str], default: str = None) -> str:
    """
    Determine which batch scheduler to use.

    This function checks a predefined list of batch schedulers in a specific order
    to determine which one is available for use. If none of the schedulers are found,
    it checks the system type environment variable to suggest a default scheduler.
    If no suitable scheduler is determined, it returns the specified default value.

    Args:
        scheduler_legend: A dictionary storing information related to each
            scheduler, including commands and expected outputs for checking their
            availability. See [`construct_scheduler_legend`][study.batch.construct_scheduler_legend]
            for more information on all the settings this dict contains.
        default: The default batch scheduler to use if a scheduler cannot be determined.

    Returns:
        The name of the available batch scheduler. Possible options include
            'slurm', 'flux', 'lsf', or 'pbs'. If no scheduler is found, returns
            the specified default value.
    """
    # These schedulers are listed in order of which should be checked for first
    # 1. Flux should be checked first due to slurm emulation scripts
    # 2. PBS should be checked before slurm for testing
    # 3. LSF should be checked before slurm for testing
    # 4. Slurm should be checked last
    schedulers_to_check = ["flux", "pbs", "lsf", "slurm"]
    for scheduler in schedulers_to_check:
        LOG.debug(f"check for {scheduler} = {check_for_scheduler(scheduler, scheduler_legend)}")
        if check_for_scheduler(scheduler, scheduler_legend):
            return scheduler

    SYS_TYPE = os.environ.get("SYS_TYPE", "")  # pylint: disable=C0103
    if "toss_3" in SYS_TYPE:
        return "slurm"

    if "blueos" in SYS_TYPE:
        return "lsf"

    return default


def get_node_count(parsed_batch: Dict, default: int = 1) -> int:
    """
    Determine a default node count based on the environment.

    This function checks the environment and the Flux version to determine the
    appropriate number of nodes to use for batch processing. It first verifies
    the Flux version, then attempts to retrieve the node count from the Flux
    allocation or environment variables specific to Slurm or LSF. If no valid
    node count can be determined, it returns a specified default value.

    Args:
        parsed_batch: A dictionary containing parsed batch configurations.
            See [`parse_batch_block`][study.batch.parse_batch_block] for more
            information on all the settings in this dictionary.
        default: The number of nodes to return if a node count from the
            environment cannot be determined.

    Returns:
        The number of nodes to use for the batch job. This value is determined
            based on the environment and scheduler specifics.

    Raises:
        ValueError: If the Flux version is too old (below 0.17.0).
    """

    # Flux version check
    flux_ver = get_flux_version(parsed_batch["flux exe"], no_errors=True)
    major, minor, _ = map(int, flux_ver.split("."))
    if major < 1 and minor < 17:
        raise ValueError("Flux version is too old. Supported versions are 0.17.0+.")

    # If flux is the scheduler, we can get the size of the allocation with this
    try:
        get_size_proc = subprocess.run("flux getattr size", shell=True, capture_output=True, text=True)
        return int(get_size_proc.stdout)
    except Exception:
        pass

    if "SLURM_JOB_NUM_NODES" in os.environ:
        return int(os.environ["SLURM_JOB_NUM_NODES"])

    # LSB systems reserve one node for launching
    if "LSB_HOSTS" in os.environ:
        nodes = set(os.environ["LSB_HOSTS"].split())
        n_batch_nodes = len(nodes) - 1
        return n_batch_nodes
    if "LSB_MCPU_HOSTS" in os.environ:
        nodes = os.environ["LSB_MCPU_HOSTS"].split()
        n_batch_nodes = len(nodes) // 2 - 1
        return n_batch_nodes

    return default


def parse_batch_block(batch: Dict) -> Dict:
    """
    Parse the batch block of a YAML configuration file.

    This function extracts relevant information from the provided batch block
    dictionary, including paths, execution options, and defaults. It retrieves
    the Flux executable path and allocation details, and populates a dictionary
    with the parsed values.

    Args:
        batch: A dictionary representing the batch block from the YAML
            configuration file.

    Returns:
        A dictionary containing parsed information from the batch block,
            including:\n
            - `btype`: The type of batch job (default is 'local').
            - `nodes`: The number of nodes to use (default is None).
            - `shell`: The shell to use (default is 'bash').
            - `bank`: The bank to charge for the job (default is an empty string).
            - `queue`: The queue to submit the job to (default is an empty string).
            - `walltime`: The maximum wall time for the job (default is an empty string).
            - `launch pre`: Any commands to run before launching (default is an empty string).
            - `launch args`: Arguments for the launch command (default is an empty string).
            - `launch command`: Custom command to launch workers. This will override the
                default launch command (default is an empty string).
            - `flux path`: Optional path to flux bin.
            - `flux exe`: The full path to the Flux executable.
            - `flux exec`: Optional flux exec command to launch workers on all nodes if
                `flux_exec_workers` is True (default is None).
            - `flux alloc`: The Flux allocation retrieved from the executable.
            - `flux opts`: Optional flux start options (default is an empty string).
            - `flux exec workers`: Optional flux argument to launch workers
                on all nodes (default is True).
    """
    flux_path: str = get_yaml_var(batch, "flux_path", "")
    if "/" in flux_path:
        flux_path += "/"

    flux_exe: str = os.path.join(flux_path, "flux")
    flux_alloc: str
    try:
        flux_alloc = get_flux_alloc(flux_exe)
    except FileNotFoundError as e:  # pylint: disable=C0103
        LOG.debug(e)
        flux_alloc = ""

    parsed_batch = {
        "btype": get_yaml_var(batch, "type", "local"),
        "nodes": get_yaml_var(batch, "nodes", None),
        "shell": get_yaml_var(batch, "shell", "bash"),
        "bank": get_yaml_var(batch, "bank", ""),
        "queue": get_yaml_var(batch, "queue", ""),
        "walltime": get_yaml_var(batch, "walltime", ""),
        "launch pre": get_yaml_var(batch, "launch_pre", ""),
        "launch args": get_yaml_var(batch, "launch_args", ""),
        "launch command": get_yaml_var(batch, "worker_launch", ""),
        "flux path": flux_path,
        "flux exe": flux_exe,
        "flux exec": get_yaml_var(batch, "flux_exec", None),
        "flux alloc": flux_alloc,
        "flux opts": get_yaml_var(batch, "flux_start_opts", ""),
        "flux exec workers": get_yaml_var(batch, "flux_exec_workers", True),
    }
    return parsed_batch


def get_flux_launch(parsed_batch: Dict) -> str:
    """
    Build the Flux launch command based on the batch section of the YAML configuration.

    This function constructs the command to launch a Flux job using the parameters
    specified in the parsed batch configuration. It determines the appropriate
    execution command for Flux workers and integrates it with the launch command
    provided in the batch configuration.

    Args:
        parsed_batch: A dictionary containing batch configuration parameters.
            See [`parse_batch_block`][study.batch.parse_batch_block] for more information
            on all the settings in this dictionary.

    Returns:
        The constructed Flux launch command, ready to be executed.
    """
    default_flux_exec = "flux exec" if parsed_batch["launch command"] else f"{parsed_batch['flux exe']} exec"
    flux_exec: str = ""
    if parsed_batch["flux exec workers"]:
        flux_exec = parsed_batch["flux exec"] if parsed_batch["flux exec"] else default_flux_exec

    if parsed_batch["launch command"] and "flux" not in parsed_batch["launch command"]:
        launch: str = (
            f"{parsed_batch['launch command']} {parsed_batch['flux exe']}"
            f" start {parsed_batch['flux opts']} {flux_exec} `which {parsed_batch['shell']}` -c"
        )
    else:
        launch: str = f"{parsed_batch['launch command']} {flux_exec} `which {parsed_batch['shell']}` -c"

    return launch


def batch_worker_launch(
    spec: MerlinSpec,
    com: str,
    nodes: Union[str, int] = None,
    batch: Dict = None,
) -> str:
    """
    Create the worker launch command based on the batch configuration in the
    workflow specification.

    This function constructs a command to launch a worker process using the
    specified batch configuration. It handles different batch types and
    integrates any necessary pre-launch commands, launch arguments, and
    node specifications.

    Args:
        spec (spec.specification.MerlinSpec): An instance of the
            [`MerlinSpec`][spec.specification.MerlinSpec] class that contains the
            configuration details, including the batch section.
        com: The command to launch with the batch configuration.
        nodes: The number of nodes to use in the batch launch. If not specified,
            it will default to the value in the batch configuration.
        batch: An optional batch override from the worker configuration. If not
            provided, the function will attempt to retrieve the batch section from
            the specification.

    Returns:
        The constructed worker launch command, ready to be executed.

    Raises:
        AttributeError: If the batch section is missing in the specification.
        TypeError: If the `nodes` parameter is of an invalid type.
    """
    if batch is None:
        try:
            batch = spec.batch
        except AttributeError:
            LOG.error("The batch section is required in the specification file.")
            raise

    parsed_batch = parse_batch_block(batch)

    # A jsrun submission cannot be run under a parent jsrun so
    # all non flux lsf submissions need to be local.
    if parsed_batch["btype"] == "local" or "lsf" in parsed_batch["btype"]:
        return com

    if nodes is None:
        # Use the value in the batch section
        nodes = parsed_batch["nodes"]

    # Get the number of nodes from the environment if unset
    if nodes is None or nodes == "all":
        nodes = get_node_count(parsed_batch, default=1)
    elif not isinstance(nodes, int):
        raise TypeError("Nodes was passed into batch_worker_launch with an invalid type (likely a string other than 'all').")

    if not parsed_batch["launch command"]:
        parsed_batch["launch command"] = construct_worker_launch_command(parsed_batch, nodes)

    if parsed_batch["launch args"]:
        parsed_batch["launch command"] += f" {parsed_batch['launch args']}"

    # Allow for any pre launch manipulation, e.g. module load
    # hwloc/1.11.10-cuda
    if parsed_batch["launch pre"]:
        parsed_batch["launch command"] = f"{parsed_batch['launch pre']} {parsed_batch['launch command']}"

    LOG.debug(f"launch command: {parsed_batch['launch command']}")

    worker_cmd: str = ""
    if parsed_batch["btype"] == "flux":
        launch = get_flux_launch(parsed_batch)
        worker_cmd = f'{launch} "{com}"'
    else:
        worker_cmd = f"{parsed_batch['launch command']} {com}"

    return worker_cmd


def construct_scheduler_legend(parsed_batch: Dict, nodes: int) -> Dict:
    """
    Constructs a legend of relevant information needed for each scheduler.

    This function generates a dictionary containing configuration details for various
    job schedulers based on the provided batch configuration. The returned dictionary
    includes flags for bank, queue, and walltime, as well as commands to check the
    scheduler and the initial launch command.

    Args:
        parsed_batch: A dictionary of batch configurations, which must include `bank`,
            `queue`, `walltime`, and `flux alloc`. See
            [`parse_batch_block`][study.batch.parse_batch_block] for more information on
            all the settings in this dictionary.
        nodes: The number of nodes to use in the launch command.

    Returns:
        A dictionary containing scheduler-related information, structured as
            follows:\n
            - For each scheduler (e.g., 'flux', 'lsf', 'pbs', 'slurm'):\n
                - `bank` (str): The flag to add a bank to the launch command.
                - `check cmd` (List[str]): The command to run to check if this is the main
                    scheduler for the cluster.
                - `expected check output` (bytes): The expected output from running
                    the check command.
                - `launch` (str): The initial launch command for the scheduler.
                - `queue` (str): The flag to add a queue to the launch command (if
                    applicable).
                - `walltime` (str): The flag to add a walltime to the launch command
                    (if applicable).
    """
    scheduler_legend = {
        "flux": {
            "bank": f" --setattr=system.bank={parsed_batch['bank']}",
            "check cmd": ["flux", "resource", "info"],
            "expected check output": b"Nodes",
            "launch": f"{parsed_batch['flux alloc']} -o pty -N {nodes} --exclusive --job-name=merlin",
            "queue": f" --setattr=system.queue={parsed_batch['queue']}",
            "walltime": f" -t {convert_timestring(parsed_batch['walltime'], format_method='FSD')}",
        },
        "lsf": {
            "check cmd": ["jsrun", "--help"],
            "expected check output": b"jsrun",
            "launch": f"jsrun -a 1 -c ALL_CPUS -g ALL_GPUS --bind=none -n {nodes}",
        },
        # pbs is mainly a placeholder in case a user wants to try it (we don't have it at the lab so it's mostly untested)
        "pbs": {
            "bank": f" -A {parsed_batch['bank']}",
            "check cmd": ["qsub", "--version"],
            "expected check output": b"pbs_version",
            "launch": f"qsub -l nodes={nodes}",
            "queue": f" -q {parsed_batch['queue']}",
            "walltime": f" -l walltime={convert_timestring(parsed_batch['walltime'])}",
        },
        "slurm": {
            "bank": f" -A {parsed_batch['bank']}",
            "check cmd": ["sbatch", "--help"],
            "expected check output": b"sbatch",
            "launch": f"srun -N {nodes} -n {nodes}",
            "queue": f" -p {parsed_batch['queue']}",
            "walltime": f" -t {convert_timestring(parsed_batch['walltime'])}",
        },
    }
    return scheduler_legend


def construct_worker_launch_command(parsed_batch: Dict, nodes: int) -> str:
    """
    Constructs the worker launch command based on the provided batch configuration.

    This function generates a launch command for a worker process when no
    'worker_launch' command is specified in the batch configuration. It
    utilizes the scheduler legend to incorporate necessary flags such as
    bank, queue, and walltime, depending on the workload manager.

    Args:
        parsed_batch: A dictionary of batch configurations, which must include
            `btype`, `bank`, `queue`, and `walltime`. See
            [`parse_batch_block`][study.batch.parse_batch_block] for more information
            on all the settings in this dictionary.
        nodes: The number of nodes to use in the batch launch.

    Returns:
        The constructed launch command for the worker process.

    Raises:
        TypeError: If the PBS scheduler is enabled for a batch type other than 'flux'.
        KeyError: If the workload manager is not found in the scheduler legend.
    """
    # Initialize launch_command and get the scheduler_legend and workload_manager
    launch_command: str = ""
    scheduler_legend: Dict = construct_scheduler_legend(parsed_batch, nodes)
    workload_manager: str = get_batch_type(scheduler_legend)

    LOG.debug(f"parsed_batch: {parsed_batch}")

    if parsed_batch["btype"] == "pbs" and workload_manager == parsed_batch["btype"]:
        raise TypeError("The PBS scheduler is only enabled for 'batch: flux' type")

    if parsed_batch["btype"] == "slurm" and workload_manager not in ("lsf", "flux", "pbs"):
        workload_manager = "slurm"

    LOG.debug(f"workload_manager: {workload_manager}")

    try:
        launch_command = scheduler_legend[workload_manager]["launch"]
    except KeyError as e:  # pylint: disable=C0103
        LOG.debug(e)

    # If lsf is the workload manager we stop here (no need to add bank, queue, walltime)
    if workload_manager != "lsf" or not launch_command:
        # Add bank, queue, and walltime to the launch command as necessary
        for key in ("bank", "queue", "walltime"):
            if parsed_batch[key]:
                try:
                    launch_command += scheduler_legend[workload_manager][key]
                except KeyError as e:  # pylint: disable=C0103
                    LOG.error(e)

        # To read from stdin we append this to the launch command for pbs
        if workload_manager == "pbs":
            launch_command += " --"

    return launch_command



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
CLI module for purging tasks from Merlin queues on the task server.

This module defines the `PurgeCommand` class, which implements the `purge`
subcommand in the Merlin CLI. The command is used to remove tasks from
queues either entirely or selectively, based on the specified steps in
a Merlin YAML workflow specification.
"""

# pylint: disable=duplicate-code

import logging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace

from merlin.ascii_art import banner_small
from merlin.cli.commands.command_entry_point import CommandEntryPoint
from merlin.cli.utils import get_merlin_spec_with_override
from merlin.router import purge_tasks


LOG = logging.getLogger("merlin")


class PurgeCommand(CommandEntryPoint):
    """
    Handles `purge` CLI command for removing tasks from queues on the server.

    Methods:
        add_parser: Adds the `purge` command to the CLI parser.
        process_command: Processes the CLI input and dispatches the appropriate action.
    """

    def add_parser(self, subparsers: ArgumentParser):
        """
        Add the `purge` command parser to the CLI argument parser.

        Parameters:
            subparsers (ArgumentParser): The subparsers object to which the `purge` command parser will be added.
        """
        purge: ArgumentParser = subparsers.add_parser(
            "purge",
            help="Remove all tasks from all merlin queues (default).              "
            "If a user would like to purge only selected queues use:    "
            "--steps to give a steplist, the queues will be defined from the step list",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        purge.set_defaults(func=self.process_command)
        purge.add_argument("specification", type=str, help="Path to a Merlin YAML spec file")
        purge.add_argument(
            "-f",
            "--force",
            action="store_true",
            dest="purge_force",
            default=False,
            help="Purge the tasks without confirmation",
        )
        purge.add_argument(
            "--steps",
            nargs="+",
            type=str,
            dest="purge_steps",
            default=["all"],
            help="The specific steps in the YAML file from which you want to purge the queues. \
            The input is a space separated list.",
        )
        purge.add_argument(  # pylint: disable=duplicate-code
            "--vars",
            action="store",
            dest="variables",
            type=str,
            nargs="+",
            default=None,
            help="Specify desired Merlin variable values to override those found in the specification. Space-delimited. "
            "Example: '--vars MY_QUEUE=hello'",
        )

    def process_command(self, args: Namespace):
        """
        CLI command for purging tasks from the task server.

        This function removes specified tasks from the task server based on the provided
        Merlin specification. It allows for targeted purging or forced removal of tasks.

        Args:
            args: Parsed CLI arguments containing:\n
                - `purge_force`: If True, forces the purge operation without confirmation.
                - `purge_steps`: Steps or criteria based on which tasks will be purged.
        """
        print(banner_small)
        spec, _ = get_merlin_spec_with_override(args)
        ret = purge_tasks(
            spec.merlin["resources"]["task_server"],
            spec,
            args.purge_force,
            args.purge_steps,
        )

        LOG.info(f"Purge return = {ret} .")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `study` package contains functionality for defining, managing, and monitoring studies
in Merlin. A study represents a collection of tasks, steps, and workflows that can be
executed on distributed systems using various schedulers.

Modules:
    batch.py: Parses the batch section of the YAML specification, supporting worker
        launches for schedulers like Slurm, LSF, and Flux.
    celeryadapter.py: Provides an adapter for integrating with the Celery Distributed
        Task Queue, enabling distributed task execution.
    dag.py: Defines the Merlin `DAG` class, which represents the Directed Acyclic Graph
        structure of a study's workflow.
    script_adapter.py: Contains functionality for adapting bash scripts to work with
        supported schedulers, including Flux, LSF, and Slurm.
    status_constants.py: Defines constants used by the `status` module and its renderers,
        helping to avoid circular import issues.
    status_renderers.py: Handles the creation of formatted, task-by-task status displays
        for studies.
    status.py: Implements functionality for retrieving and displaying the statuses of studies.
    step.py: Contains the logic for representing and managing individual steps in a study.
    study.py: Implements the core logic for defining and managing a study as a whole.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This file contains all of the constants used for the status command.
Separating this from [`status.py`][study.status] and [`status_renderers.py`][study.status_renderers]
helps with circular import issues.
"""

VALID_STATUS_FILTERS = ("INITIALIZED", "RUNNING", "FINISHED", "FAILED", "CANCELLED", "DRY_RUN", "UNKNOWN")
VALID_RETURN_CODES = ("SUCCESS", "SOFT_FAIL", "HARD_FAIL", "STOP_WORKERS", "RESTART", "RETRY", "DRY_SUCCESS", "UNRECOGNIZED")
VALID_EXIT_FILTERS = ("E", "EXIT")
ALL_VALID_FILTERS = VALID_STATUS_FILTERS + VALID_RETURN_CODES + VALID_EXIT_FILTERS + ("MAX_TASKS",)

# Listing worker_name below since it was used in v1.12.0 so if you try to run "merlin status" on a study
# ran with 1.12.0, then you'll need this key here for everything to function
CELERY_KEYS = set(["task_queue", "workers", "worker_name"])
RUN_TIME_STAT_KEYS = set(["avg_run_time", "run_time_std_dev"])
NON_WORKSPACE_KEYS = CELERY_KEYS.union(RUN_TIME_STAT_KEYS)
NON_WORKSPACE_KEYS.add("parameters")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Main entry point into Merlin's codebase.
"""

import logging
import sys
import traceback

from merlin.cli.argparse_main import build_main_parser
from merlin.log_formatter import setup_logging


LOG = logging.getLogger("merlin")


def main():
    """
    Entry point for the Merlin command-line interface (CLI) operations.

    This function sets up the argument parser, handles command-line arguments,
    initializes logging, and executes the appropriate function based on the
    provided command. It ensures that the user receives help information if
    no arguments are provided and performs error handling for any exceptions
    that may occur during command execution.
    """
    parser = build_main_parser()
    if len(sys.argv) == 1:
        parser.print_help(sys.stdout)
        return 1
    args = parser.parse_args()

    setup_logging(logger=LOG, log_level=args.level.upper(), colors=True)

    try:
        args.func(args)
        # pylint complains that this exception is too broad - being at the literal top of the program stack, it's ok.
    except Exception as excpt:  # pylint: disable=broad-except
        LOG.debug(traceback.format_exc())
        LOG.error(str(excpt))
        sys.exit(1)

    # All paths in a function ought to return an exit code, or none of them should. Given the
    # distributed nature of Merlin, maybe it doesn't make sense for it to exit 0 until the work is completed, but
    # if the work is dispatched with no errors, that is a 'successful' Merlin run - any other failures are runtime.
    sys.exit()


if __name__ == "__main__":
    main()



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `server` package defines the functionality for managing a containerized server
through Merlin.

Modules:
    server_commands.py: Main functions to interact with the server.
    server_config.py: Server configuration functions.
    server_util.py: Defines the structure of our server configurations.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Base Classes for Redis-Backed Data Stores in Merlin

This module provides foundational classes for implementing Redis-based persistence
in the Merlin system. It defines reusable components that standardize how data models
are stored, retrieved, and managed in Redis.

See also:
    - merlin.backends.store_base: Base class
    - merlin.backends.redis.redis_stores: Concrete store implementations
    - merlin.db_scripts.data_models: Data model definitions
"""

import logging
from typing import Generic, List, Optional, Type

from redis import Redis

from merlin.backends.store_base import StoreBase, T
from merlin.backends.utils import deserialize_entity, get_not_found_error_class, serialize_entity


LOG = logging.getLogger(__name__)


class RedisStoreBase(StoreBase[T], Generic[T]):
    """
    Base class for Redis-based stores.

    This class provides common functionality for saving, retrieving, and deleting
    entities in a Redis database.

    Attributes:
        client (Redis): The Redis client used for database operations.
        key (str): The prefix key used for Redis entries.
        model_class (Type[T]): The model class used for deserialization.

    Methods:
        save: Save or update an entity in the database.
        retrieve: Retrieve an entity from the database by ID.
        retrieve_all: Query the database for all entities of this type.
        delete: Delete an entity from the database by ID.
    """

    def __init__(self, client: Redis, key: str, model_class: Type[T]):
        """
        Initialize the Redis store with a Redis client.

        Args:
            client: A Redis client instance used to interact with the Redis database.
            key: The prefix key used for Redis entries.
            model_class: The model class used for deserialization.
        """
        self.client: Redis = client
        self.key: str = key
        self.model_class: Type[T] = model_class

    def _get_full_key(self, entity_id: str) -> str:
        """
        Get the full Redis key for an entity.

        Args:
            entity_id: The entity ID.

        Returns:
            The full Redis key.
        """
        return entity_id if entity_id.startswith(f"{self.key}:") else f"{self.key}:{entity_id}"

    def save(self, entity: T):
        """
        Save or update an entity in the Redis database.

        Args:
            entity: The entity to save.
        """
        entity_key = f"{self.key}:{entity.id}"

        if self.client.exists(entity_key):
            LOG.debug(f"Attempting to update {self.key} with id '{entity.id}'...")
            # Get the existing data from Redis and convert it to an instance of BaseDataModel
            existing_data = self.client.hgetall(entity_key)
            existing_data_class = deserialize_entity(existing_data, self.model_class)

            # Update the fields and save it to Redis
            existing_data_class.update_fields(entity.to_dict())
            updated_data = serialize_entity(existing_data_class)
            self.client.hset(entity_key, mapping=updated_data)
            LOG.debug(f"Successfully updated {self.key} with id '{entity.id}'.")
        else:
            LOG.debug(f"Creating a {self.key} entry in Redis...")
            serialized_data = serialize_entity(entity)
            self.client.hset(entity_key, mapping=serialized_data)
            LOG.debug(f"Successfully created a {self.key} with id '{entity.id}' in Redis.")

    def retrieve(self, identifier: str) -> Optional[T]:
        """
        Retrieve an entity from the Redis database by ID.

        Args:
            identifier: The ID of the entity to retrieve.

        Returns:
            The entity if found, None otherwise.
        """
        LOG.debug(f"Retrieving identifier {identifier} in RedisStoreBase.")
        entity_key = self._get_full_key(identifier)
        if not self.client.exists(entity_key):
            return None

        data_from_redis = self.client.hgetall(entity_key)
        return deserialize_entity(data_from_redis, self.model_class)

    def retrieve_all(self) -> List[T]:
        """
        Query the Redis database for all entities of this type.

        Returns:
            A list of entities.
        """
        entity_type = f"{self.key}s" if self.key != "study" else "studies"
        LOG.info(f"Fetching all {entity_type} from Redis...")

        pattern = f"{self.key}:*"
        all_entities = []

        # Exclude name mapping key if it exists
        keys_to_exclude = {f"{self.key}:name"}

        # Loop through all entities using scan_iter for better efficiency with large datasets
        for key in self.client.scan_iter(match=pattern):
            if key in keys_to_exclude:
                continue

            entity_id = key.split(":")[1]  # Extract the ID for logging
            try:
                entity_info = self.retrieve(key)
                if entity_info:
                    all_entities.append(entity_info)
                else:
                    LOG.warning(f"{self.key.capitalize()} with id '{entity_id}' could not be retrieved or does not exist.")
            except Exception as exc:  # pylint: disable=broad-except
                LOG.error(f"Error retrieving {self.key} with id '{entity_id}': {exc}")

        LOG.info(f"Successfully retrieved {len(all_entities)} {entity_type} from Redis.")
        return all_entities

    def delete(self, identifier: str):
        """
        Delete an entity from the Redis database by ID.

        Args:
            identifier: The ID of the entity to delete.
        """
        LOG.info(f"Attempting to delete {self.key} with id '{identifier}' from Redis...")

        entity = self.retrieve(identifier)
        if entity is None:
            error_class = get_not_found_error_class(self.model_class)
            raise error_class(f"{self.key.capitalize()} with id '{identifier}' does not exist in the database.")

        # Delete the entity's hash entry
        entity_key = f"{self.key}:{entity.id}"
        LOG.debug(f"Deleting {self.key} hash with key '{entity_key}'...")
        self.client.delete(entity_key)
        LOG.debug(f"Successfully removed {self.key} hash from Redis.")

        LOG.info(f"Successfully deleted {self.key} '{identifier}' from Redis.")


class NameMappingMixin:
    """
    Mixin class that adds name-to-ID mapping functionality to Redis stores.

    This mixin extends Redis stores to support retrieval and deletion by name.

    Methods:
        save: Save or update an entity in the database.
        retrieve: Retrieve an entity from the database by ID or name.
        delete: Delete an entity from the database by ID or name.
    """

    def save(self, entity: Type[T]):
        """
        Save an entity and update the name-to-ID mapping.

        Args:
            entity: The entity to save.
        """
        name_or_ws = entity.name if hasattr(entity, "name") else entity.workspace
        LOG.debug(f"Saving entity {name_or_ws} with id {entity.id} in NameMappingMixin.")
        existing_entity_id = self.client.hget(f"{self.key}:name", name_or_ws)

        # Call the parent class's save method
        super().save(entity)

        # Update name-to-ID mapping if it's a new entity
        if not existing_entity_id:
            LOG.debug(f"Creating a new name-to-ID mapping for {name_or_ws} with id {entity.id}")
            self.client.hset(f"{self.key}:name", name_or_ws, entity.id)

    def retrieve(self, identifier: str, by_name: bool = False) -> Optional[T]:
        """
        Retrieve an entity from the Redis database, either by ID or name.

        Args:
            identifier: The ID or name of the entity to retrieve.
            by_name: If True, interpret the identifier as a name. If False, interpret it as an ID.

        Returns:
            The entity if found, None otherwise.
        """
        LOG.debug(f"Retrieving identifier {identifier} in NameMappingMixin.")
        if by_name:
            # Retrieve the entity ID using the name-to-ID mapping
            entity_id = self.client.hget(f"{self.key}:name", identifier)
            if entity_id is None:
                LOG.debug("Could not retrieve entity id by name-to-ID mapping.")
                return None
            return super().retrieve(entity_id)
        # Use the parent class's retrieve method for ID-based retrieval
        return super().retrieve(identifier)

    def delete(self, identifier: str, by_name: bool = False):
        """
        Delete an entity from the Redis database, either by ID or name.

        Args:
            identifier: The ID or name of the entity to delete.
            by_name: If True, interpret the identifier as a name. If False, interpret it as an ID.
        """
        id_type = "name" if by_name else "id"
        LOG.debug(f"Attempting to delete {self.key} with {id_type} '{identifier}' from Redis...")

        # Retrieve the entity to ensure it exists and get its ID and name
        entity = self.retrieve(identifier, by_name=by_name)
        if entity is None:
            error_class = get_not_found_error_class(self.model_class)
            raise error_class(f"{self.key.capitalize()} with {id_type} '{identifier}' not found in the database.")

        # Delete the entity from the name index and Redis
        name_or_ws = entity.name if hasattr(entity, "name") else entity.workspace
        self.client.hdel(f"{self.key}:name", name_or_ws)
        self.client.delete(f"{self.key}:{entity.id}")

        LOG.info(f"Successfully deleted {self.key} with {id_type} '{identifier}'.")



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module contains a list of examples that can be used when learning to use
Merlin, or for setting up new workflows.

Examples are packaged in directories, with the directory name denoting
the example name. This must match the name of the Merlin specification inside.
"""
import glob
import logging
import os
import shutil
from typing import Dict, List, Union

import tabulate
import yaml


LOG = logging.getLogger(__name__)

EXAMPLES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "workflows")

# TODO modify the example command to eliminate redundancy
# - e.g. running `merlin example flux_local` will produce the same output
#   as running `merlin example flux_par` or `merlin example flux_par_restart`.
#   This should just be `merlin example flux`.
# - restart and restart delay should be one example
# - feature demo and remote feature demo should be one example
# - all openfoam examples should just be under one openfoam label


def gather_example_dirs() -> Dict[str, str]:
    """
    Get all the example directories.

    Returns:
        A dictionary where the keys and values are the names of example directories.
    """
    result = {}
    for directory in sorted(os.listdir(EXAMPLES_DIR)):
        result[directory] = directory
    return result


def gather_all_examples() -> List[str]:
    """
    Get all the example YAML files.

    Returns:
        A list of file paths to all YAML files in the example directories.
    """
    path = os.path.join(os.path.join(EXAMPLES_DIR, ""), os.path.join("*", "*.yaml"))
    return glob.glob(path)


def write_example(src_path: str, dst_path: str):
    """
    Write out the example workflow to a file or directory.

    Args:
        src_path: The path to copy the example from.
        dst_path: The destination path to copy the example to.
    """
    if os.path.isdir(src_path):
        shutil.copytree(src_path, dst_path)
    else:
        shutil.copy(src_path, dst_path)


def list_examples() -> str:
    """
    List all available examples with their descriptions.

    Returns:
        A formatted string table of example names and their descriptions.
    """
    headers = ["name", "description"]
    rows = []
    for example_dir in gather_example_dirs():
        directory = os.path.join(os.path.join(EXAMPLES_DIR, example_dir), "")
        specs = glob.glob(directory + "*.yaml")
        for spec in sorted(specs):
            if "template" in spec:
                continue
            with open(spec) as f:  # pylint: disable=C0103
                try:
                    spec_metadata = yaml.safe_load(f)["description"]
                except KeyError:
                    LOG.warning(f"{spec} lacks required section 'description'")
                    continue
                except TypeError:
                    continue
            name = spec_metadata["name"]
            if name is None:
                continue
            # if there is a variable reference in the workflow name, instead list the filename (minus the yaml extension).
            if "$" in name:
                name = os.path.basename(os.path.normpath(spec)).replace(".yaml", "")
            rows.append([name, spec_metadata["description"]])
    return "\n" + tabulate.tabulate(rows, headers) + "\n"


def setup_example(name: str, outdir: str) -> Union[str, None]:
    """
    Set up the given example by copying it to the specified output directory.

    Args:
        name: The name of the example to set up.
        outdir: The output directory where the example will be copied.

    Returns:
        The name of the example if successful, or None if the example was not found or an error occurred.
    """
    example = None
    spec_paths = gather_all_examples()
    spec_path = None
    for spec_path in spec_paths:
        spec = os.path.basename(os.path.normpath(spec_path)).replace(".yaml", "")
        if name == spec:
            example = os.path.basename(os.path.dirname(spec_path))
            break
    if example is None:
        LOG.error(f"Example '{name}' not found.")
        return None

    # if there is only 1 file in the example, don't bother making a directory for it
    if len(os.listdir(os.path.dirname(spec_path))) == 1:
        src_path = os.path.join(EXAMPLES_DIR, os.path.join(example, example + ".yaml"))

    else:
        src_path = os.path.join(EXAMPLES_DIR, example)

        if outdir:
            outdir = os.path.join(os.getcwd(), outdir)
        else:
            outdir = os.path.join(os.getcwd(), example)

        if os.path.exists(outdir):
            LOG.error(f"File '{outdir}' already exists!")
            return None

    if outdir is None:
        outdir = os.getcwd()

    LOG.info(f"Copying example '{name}' to {outdir}")
    write_example(src_path, outdir)
    return example



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `entity_managers` package contains classes responsible for managing high-level database
operations across all entity types in the Merlin system.

Each manager provides CRUD (Create, Read, Update, Delete) operations tailored to a specific
entity, such as studies, runs, or workers. These managers act as intermediaries between the
underlying data models and higher-level business logic, encapsulating shared behaviors and
coordinating entity relationships, lifecycle events, and cross-references.

Modules:
    entity_manager.py: Defines the abstract base class
        [`EntityManager`][db_scripts.entity_managers.entity_manager.EntityManager],
        which provides shared infrastructure and helper methods for all entity managers.
    logical_worker_manager.py: Manages logical workers, including queue-based identity
        resolution and cleanup of run associations on deletion.
    physical_worker_manager.py: Manages physical workers and handles cleanup of references
        from associated logical workers.
    run_manager.py: Manages run entities and coordinates updates to related studies and
        logical workers.
    study_manager.py: Manages study entities and orchestrates the creation and deletion of
        studies along with their associated runs.

These managers rely on the Merlin results backend and may optionally reference each other
via the central `MerlinDatabase` class to perform coordinated, entity-spanning operations.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `db_scripts` package provides the core database infrastructure for the Merlin system.

This package defines the models, entities, and management logic required to persist, retrieve,
and manipulate data stored in Merlin's database. It encapsulates both low-level data models and
high-level entity managers, offering a structured, maintainable interface for all database
interactions within the system.

Subpackages:
    - `entities/`: Contains entity classes that wrap database models and expose higher-level
      operations such as `save`, `delete`, and `reload`.
    - `entity_managers/`: Provides manager classes that orchestrate CRUD operations across
      entity types, with support for inter-entity references and cleanup routines.

Modules:
    data_models.py: Defines the dataclasses used to represent raw records in the database,
        such as [`StudyModel`][db_scripts.data_models.StudyModel], [`RunModel`][db_scripts.data_models.RunModel],
        etc.
    db_commands.py: Exposes database-related commands intended for external use (e.g., CLI or scripts),
        allowing users to interact with Merlin's stored data.
    merlin_db.py: Contains the [`MerlinDatabase`][db_scripts.merlin_db.MerlinDatabase] class, which
        aggregates all entity managers and acts as the central access point for database operations
        across the system.
"""



import yaml
from jinja2 import Environment, FileSystemLoader, meta


# get all variable in template file
def get_variables(filename):
    env = Environment(loader=FileSystemLoader("./"))
    template_source = env.loader.get_source(env, filename)[0]
    parsed_content = env.parse(template_source)

    return meta.find_undeclared_variables(parsed_content)


def get_dict_from_yaml(filename, get_template=True):
    env = Environment(loader=FileSystemLoader("./"))
    template = env.get_template(filename)
    outputText = template.render()

    if get_template:
        return yaml.safe_load(outputText), template
    return yaml.safe_load(outputText)


def get_bounds_X(test_function):
    return {
        "rosenbrock": str([[-2, 2] for i in range(N_DIMS)]).replace(" ", ""),
        "ackley": str([[-5, 5] for i in range(2)]).replace(" ", ""),
        "rastrigin": str([[-5.12, 5.12] for i in range(N_DIMS)]).replace(" ", ""),
        "griewank": str([[-10, 10] for i in range(N_DIMS)]).replace(" ", ""),
    }[test_function]


filename = "template_optimization.temp"

workflow_dict, template = get_dict_from_yaml(filename)
undefined_variables = get_variables(filename)

TEST_FUNCTION = workflow_dict["env"]["variables"]["TEST_FUNCTION"]
DEBUG = workflow_dict["env"]["variables"]["DEBUG"]
N_DIMS = workflow_dict["env"]["variables"]["N_DIMS"]

if (TEST_FUNCTION == "ackley") and (N_DIMS != 2):
    raise Exception("The ackley function only accepts 2 dims, change the N_DIMS variable")

bounds_x = get_bounds_X(TEST_FUNCTION)
uncerts_x = [0.1 for i in range(N_DIMS)]
column_labels = [f"INPUT_{i + 1}" for i in range(N_DIMS)]

if DEBUG:
    # Reduce the number of iterations and the number of samples per iteration
    max_iter = 3

    n_samples = 2
    n_exploit = 2
    n_samples_line = 2
    n_samples_start = 8

defined_variables = {}
for undefined_variable in undefined_variables:
    if undefined_variable in locals():
        defined_variables[undefined_variable] = globals()[undefined_variable]
    else:
        print(f"Variable '{undefined_variable}' is not defined, using default specified in yaml file")

# to save the results
with open("optimization.yaml", "w") as fh:
    fh.write(template.render(defined_variables))



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Abstract base class for results backends in the Merlin application.

This module defines `ResultsBackend`, an abstract base class that specifies
the required interface for backend implementations responsible for persisting
and retrieving data in Merlin.

The `ResultsBackend` class encapsulates:
- A unified interface for saving, retrieving, and deleting Merlin data models
- Store type routing for specific model categories (study, run, logical worker, physical worker)
- Support for backend-specific configuration such as version reporting and database flushing

Usage:
    This base class is not meant to be instantiated directly. Instead, it should be subclassed
    by backend-specific implementations such as `RedisBackend` or `SQLiteBackend`.
"""

import logging
import uuid
from abc import ABC, abstractmethod
from typing import Dict, List

from merlin.backends.store_base import StoreBase
from merlin.db_scripts.data_models import BaseDataModel, LogicalWorkerModel, PhysicalWorkerModel, RunModel, StudyModel
from merlin.exceptions import UnsupportedDataModelError


LOG = logging.getLogger(__name__)


class ResultsBackend(ABC):
    """
    Abstract base class for a results backend, which provides methods to save and retrieve
    information from a backend database.

    This class defines the interface that must be implemented by any concrete backend, and serves
    as the foundation for interacting with various backend types such as Redis, PostgreSQL, etc.

    Attributes:
        backend_name (str): The name of the backend (e.g., "redis", "postgresql").
        stores (Dict[str, backends.store_base.StoreBase]): A dictionary of stores that each concrete
            implementation of this class will need to define.

    Methods:
        get_name:
            Retrieve the name of the backend.

        get_version:
            Query the backend for the current version.

        get_connection_string:
            Retrieve the connection string used to connect to the backend.

        flush_database:
            Remove every entry in the database.

        save:
            Save an entity (e.g., a study, run, or worker) to the backend database.

        retrieve:
            Retrieve an entity from the backend database using its identifier and store type.

        retrieve_all:
            Retrieve all entities of a specific type from the backend database.

        delete:
            Delete an entity from the backend database using its identifier and store type.
    """

    def __init__(self, backend_name: str, stores: Dict[str, StoreBase]):
        """
        Initialize the `ResultsBackend` instance.

        Args:
            backend_name: The name of the backend (e.g., "redis").
        """
        self.backend_name: str = backend_name
        self.stores: Dict[str, StoreBase] = stores

    def get_name(self) -> str:
        """
        Get the name of the backend.

        Returns:
            The name of the backend (e.g. redis).
        """
        return self.backend_name

    @abstractmethod
    def get_version(self) -> str:
        """
        Query the backend for the current version.

        Returns:
            A string representing the current version of the backend.
        """
        raise NotImplementedError("Subclasses of `ResultsBackend` must implement a `get_version` method.")

    def get_connection_string(self) -> str:
        """
        Query the backend for the connection string.

        Returns:
            A string representing the connection to the backend.
        """
        from merlin.config.results_backend import get_connection_string  # pylint: disable=import-outside-toplevel

        return get_connection_string(include_password=False)

    @abstractmethod
    def flush_database(self):
        """
        Remove everything stored in the database.
        """
        raise NotImplementedError("Subclasses of `ResultsBackend` must implement a `flush_database` method.")

    def _get_store_by_type(self, store_type: str) -> StoreBase:
        """
        Get the appropriate store based on the store type.

        Args:
            store_type (str): The type of store.

        Returns:
            The corresponding store.

        Raises:
            ValueError: If the `store_type` is invalid.
        """
        if store_type not in self.stores:
            raise ValueError(f"Invalid store type '{store_type}'.")
        return self.stores[store_type]

    def _get_store_by_entity(self, entity: BaseDataModel) -> StoreBase:
        """
        Get the appropriate store based on the entity type.

        Args:
            entity (BaseDataModel): The entity to save.

        Returns:
            RedisStore: The corresponding store.

        Raises:
            UnsupportedDataModelError: If the entity type is unsupported.
        """
        if isinstance(entity, StudyModel):
            return self.stores["study"]
        if isinstance(entity, RunModel):
            return self.stores["run"]
        if isinstance(entity, LogicalWorkerModel):
            return self.stores["logical_worker"]
        if isinstance(entity, PhysicalWorkerModel):
            return self.stores["physical_worker"]
        raise UnsupportedDataModelError(f"Unsupported data model of type {type(entity)}.")

    def save(self, entity: BaseDataModel):
        """
        Save a `BaseDataModel` object to the Redis database.

        Args:
            entity (BaseDataModel): An instance of one of `BaseDataModel`'s inherited classes.

        Raises:
            UnsupportedDataModelError: If the entity type is unsupported.
        """
        store = self._get_store_by_entity(entity)
        store.save(entity)

    def retrieve(self, entity_identifier: str, store_type: str) -> BaseDataModel:
        """
        Retrieve an object from the appropriate store based on the given query identifier and store type.

        Args:
            entity_identifier (str): The identifier used to query the store, either an ID (UUID) or a name.
            store_type (str): The type of store to query. Valid options are:
                - `study`
                - `run`
                - `logical_worker`
                - `physical_worker`

        Returns:
            The object retrieved from the specified store.
        """
        LOG.debug(f"Retrieving '{entity_identifier}' from store '{store_type}'.")
        store = self._get_store_by_type(store_type)
        if store_type in ["study", "physical_worker", "run"]:
            try:
                uuid.UUID(entity_identifier)
                return store.retrieve(entity_identifier)
            except ValueError:
                return store.retrieve(entity_identifier, by_name=True)
        else:
            return store.retrieve(entity_identifier)

    def retrieve_all(self, store_type: str) -> List[BaseDataModel]:
        """
        Retrieve all objects from the specified store.

        Args:
            store_type (str): The type of store to query. Valid options are:
                - `study`
                - `run`
                - `logical_worker`
                - `physical_worker`

        Returns:
            A list of objects retrieved from the specified store.
        """
        store = self._get_store_by_type(store_type)
        return store.retrieve_all()

    def delete(self, entity_identifier: str, store_type: str):
        """
        Delete an entity from the specified store.

        Args:
            entity_identifier (str): The identifier of the entity to delete.
            store_type (str): The type of store to query. Valid options are:
                - `study`
                - `run`
                - `logical_worker`
                - `physical_worker`
        """
        store = self._get_store_by_type(store_type)
        if store_type in ["study", "physical_worker"]:
            try:
                uuid.UUID(entity_identifier)
                store.delete(entity_identifier)
            except ValueError:
                store.delete(entity_identifier, by_name=True)
        else:
            store.delete(entity_identifier)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module contains example specification files for Merlin workflows,
along with detailed explanations of each block in the specification.

The examples and templates are useful for understanding how to structure
Merlin workflows, define tasks, manage parameters, and configure resources.
"""

# Taken from https://lc.llnl.gov/mlsi/docs/merlin/merlin_config.html
TEMPLATE_FILE_CONTENTS = """
####################################
# Description Block (Required)
####################################
# The description block is where the description of the study is placed. This
# section is meant primarily for documentation purposes so that when a
# specification is passed to other users they can glean a general understanding
# of what this study is meant to achieve.
#-------------------------------
# Required keys:
#   name - Name of the study
#   description - Description of what this study does.
#-------------------------------
# NOTE: You can add other keys to this block for custom documentation. Merlin
# currently only looks for the required set.
####################################
description:
  description: Run a scan through Merlin
  name: MERLIN

####################################
# Batch Block (Required)
####################################
# The batch system to use for each allocation
#-------------------------------
# Required keys:
#   type - The scheduler type to use (local|slurm|flux|lsf)
#   bank - The allocation bank
#   queue - The batch queue
####################################
batch:
   type: flux
   bank: testbank
   queue: pbatch
   flux_path: <optional path to flux bin>
   flux_start_opts: <optional flux start options
   shell: <the interpreter to use for the script after the shebang>
          # e.g. /bin/bash, /bin/tcsh, python, /usr/bin/env perl, etc.


#####################################
# Environment Block
####################################
# The environment block is where items describing the study's environment are
# defined. This includes static information that the study needs to know about
# and dependencies that the workflow requires for execution.
#-------------------------------
# NOTE: This block isn't strictly required as a study may not depend on anything.
########################################################################
env:
  #-------------------------------
  # Variables
  #-------------------------------
  # Values that the workflow substitutes into steps and are similar in
  # concept to Unix environment variables. These variables are not dependent
  # on values in the environment and so are more portable.
  # set a custom output path for the study workspace. This path is where
  # Merlin will place all temporary files, state files, and any output.
  # The resulting path is usually a timestamped folder within OUTPUT_PATH
  # and in this case would be './sample_output/merlin/merlin_sample1_<timestamp>'.
  # Variables are useful for ensuring consistency with fixed formatting for
  # output files, or fixed formatting for components of steps.

  # NOTE: If not specified, OUTPUT_PATH is assumed to be the path where Merlin was launched from.
  # NOTE: If the '-o' flag is specified for the run subcommand, OUTPUT_PATH
  # the output path will be taken from there and will not generate a
  # timestamped path.
  #-------------------------------
  variables:
      OUTPUT_PATH: ./studies

####################################
# Study Block (Required)
####################################
# The study block is where the steps in the workflow are defined. This section
# of the specification represents the unexpanded set of tasks that the study
# is composed of.
#
#
# A description of what gets turned into tasks and what type of task
# would be a good addition
#
#######################################################################
study:
  - name: runs1
    description: Run on alloc1
    run:
     cmd: echo "$(VAR1) $(VAR2)" > simrun.out
     task_queue: queue1

  - name: post-process
    description: Post-Process runs on alloc1
    run:
      cmd: |
        cd $(runs1.workspace)/$(MERLIN_SAMPLE_PATH)
        echo "<post-process-command>"
      depends: [runs1]
      task_queue: queue1

  - name: runs2
    description: Run on alloc2
    run:
      cmd: |
        touch learnrun.out
        echo "$(VAR1) $(VAR2)" >> learnrun.out
        exit $(MERLIN_RESTART) # some syntax to catch a retry error code
      task_queue: lqueue
      max_retries: 5 # workflow will fail if retries exceeds this

  - name: monitor
    description: Monitor on alloc1
    run:
      cmd: date > monitor.out
      task_queue: mqueue

####################################
# Parameter Block (Required)
####################################
# The parameter block contains all the things we'd like to vary in the study.
# Currently, there are two modes of operating in the specification:
# 1. If a parameter block is specified, the study is expanded and considered a
#   parameterized study.
# 2. If a parameter block is not specified, the study is treated as linear and
#    the resulting study is not expanded.
#
# There are three keys per parameter:
# 1. A list of values that the parameter takes.
# 2. A label that represents a "pretty printed" version of the parameter. The
#    parameter values is specified by the '%%' moniker (for example, for SIZE --
#    when SIZE is equal to 10, the label will be 'SIZE.10'). To access the label
#    for SIZE, for example, the token '$(SIZE.label)' is used.
#    Labels can take one of two forms: A single string with the '%%' marker or
#    a list of per value labels (must be the same length as the list of values).
#
# NOTE: A specified parameter does not necessarily have to be used in every step
# or at all. If a parameter is specified and not used, it simply will not be
# factored into expansion or the naming of expanded steps or their workspaces.
# NOTE: You can also specify custom generation of parameters using a Python
# file containing the definition of a function as follows:
#
# 'def get_custom_generator():'
#
# The 'get_custom_generator' function is required to return a ParameterGenerator
# instance populated with custom filled values. In order to use the file, simply
# call Merlin using 'merlin run <specification path>'.
########################################################################
global.parameters:
  STUDY:
    label: STUDY.%%
    values: [MERLIN1, MERLIN2]
  SIZE:
    label   : SIZE.%%
    values  : [10, 20]
  ITERATIONS:
    label   : ITER.%%
    values  : [10, 20]

####################################
# Merlin Block (Required)
####################################
# The merlin-specific block will add any required configuration to the study
# DAG including task server config, data management and sample definitions.
#
# Merlin will replace all SPECROOT instances with the directory where
# the input yaml was run.
#######################################################################
merlin:

  ####################################
  # Resource definitions
  #
  # Define the task server configuration and workers to run the tasks.
  #
  ####################################
  resources:
    task_server: celery
    overlap: False
    workers:
      simworkers:
        args: # <celery worker args> <optional>
        steps: [runs1, post-process, monitor]  # [all] when steps is omitted

      learnworkers:
        args: # <celery worker args> <optional>
        steps: [runs2]

  ####################################
  # Sample definitions
  ####################################
  samples:
    column_labels: [VAR1, VAR2]
    file: $(MERLIN_INFO)/samples.csv
    generate:
      cmd: |
        echo "Generate samples here."
        echo "1.0,1.0" > $(MERLIN_INFO)/samples.csv
        echo "2.0,2.0" >> $(MERLIN_INFO)/samples.csv
"""


SIMPLE_EXAMPLE = """
description:
    name: simple_example
    description: This is a single one step Merlin workflow example.

study:
    - name: iterate_count
      description: A simple counter.
      run:
        cmd: |
            echo "Count: $(COUNT)"

global.parameters:
    COUNT:
        values : [1, 2, 3, 4]
        label  : COUNT.%%
"""


# Metadata for the template.
TEMPLATES = [
    {
        "name": "sample_template",
        "filename": "sample_template.yaml",
        "description": "A simple single step Merlin workflow example",
        "content": SIMPLE_EXAMPLE,
    },
    {
        "name": "merlin_documentation",
        "filename": "template_spec.yaml",
        "description": "A fully documented Merlin example spec example",
        "content": TEMPLATE_FILE_CONTENTS,
    },
]



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module handles the expansion of variables in a Merlin spec file.

It provides functionality to expand user-defined variables, environment variables, and reserved variables
within a spec file. The module also supports variable substitution for specific use cases, such as parameter
substitutions for samples and commands, and allows for the processing of override variables provided via
the command-line interface.
"""

import logging
from collections import ChainMap
from copy import deepcopy
from os.path import expanduser, expandvars
from typing import Dict, List, Tuple

from merlin.common.enums import ReturnCode
from merlin.spec.override import error_override_vars, replace_override_vars
from merlin.spec.specification import MerlinSpec
from merlin.utils import contains_shell_ref, contains_token, verify_filepath


MAESTRO_RESERVED = {"SPECROOT", "WORKSPACE", "LAUNCHER"}
STEP_AWARE = {
    "MERLIN_GLOB_PATH",
    "MERLIN_PATHS_ALL",
    "MERLIN_SAMPLE_ID",
    "MERLIN_SAMPLE_PATH",
}
PROVENANCE_REPLACE = {
    "MERLIN_WORKSPACE",
    "MERLIN_TIMESTAMP",
    "MERLIN_INFO",
    "MERLIN_SUCCESS",
    "MERLIN_RESTART",
    "MERLIN_SOFT_FAIL",
    "MERLIN_HARD_FAIL",
    "MERLIN_RETRY",
    "MERLIN_STOP_WORKERS",
    "MERLIN_RAISE_ERROR",
}
MERLIN_RESERVED = STEP_AWARE | PROVENANCE_REPLACE
RESERVED = MAESTRO_RESERVED | MERLIN_RESERVED


LOG = logging.getLogger(__name__)


def var_ref(string: str) -> str:
    """
    Format a string as a variable reference.

    This function takes a string, converts it to uppercase, and returns it
    wrapped in the format `$(<string>)`. If the string already contains
    a token (e.g., it is already formatted as a variable reference), a warning
    is logged and the original string is returned unchanged.

    Args:
        string: The input string to format as a variable reference.

    Returns:
        The formatted variable reference, or the original string if it
            already contains a token.

    Example:
        ```python
        >>> var_ref("example")
        '$(EXAMPLE)'
        ```
    """
    string = string.upper()
    if contains_token(string):
        LOG.warning(f"Bad var_ref usage on string '{string}'.")
        return string
    return f"$({string})"


def expand_line(line: str, var_dict: Dict[str, str], env_vars: bool = False) -> str:
    """
    Expand a single line of text by substituting variables.

    This function replaces variable references in a given line of text with
    their corresponding values from a provided dictionary. Optionally, it can
    also expand environment variables and user home directory shortcuts (e.g., `~`).

    Args:
        line: The input line of text to expand.
        var_dict: A dictionary of variable names and their corresponding values
            to substitute in the line.
        env_vars: If True, environment variables and home directory shortcuts
            will also be expanded.

    Returns:
        The expanded line of text with all applicable substitutions applied.

    Example:
        ```python
        >>> line = "Path: $(VAR1) and $(VAR2)"
        >>> var_dict = {"VAR1": "/path/to/dir", "VAR2": "/another/path"}
        >>> expand_line(line, var_dict)
        'Path: /path/to/dir and /another/path'
        ```
    """
    # fmt: off
    if (
        (not contains_token(line))
        and (not contains_shell_ref(line))
        and ("~" not in line)
    ):
        return line
    # fmt: on
    for key, val in var_dict.items():
        if key in line:
            line = line.replace(var_ref(key), str(val))
    if env_vars:
        line = expandvars(expanduser(line))
    return line


def expand_by_line(text: str, var_dict: Dict[str, str]) -> str:
    """
    Expand variables in a text line by line.

    This function processes a multi-line text (e.g., a YAML specification) and
    replaces variable references in each line using a provided dictionary of
    variable names and their corresponding values.

    Args:
        text: The input multi-line text to process.
        var_dict: A dictionary of variable names and their corresponding values
            to substitute in the text.

    Returns:
        The text with all applicable variable substitutions applied, processed
            line by line.

    Example:
        ```python
        >>> text = "Path is $(VAR1) and stores $(VAR2)"
        >>> var_dict = {"VAR1": "/path/to/dir", "VAR2": "value"}
        >>> expand_by_line(text, var_dict)
        'Path is /path/to/dir and stores value'
        ```
    """
    text = text.splitlines()
    result = ""
    for line in text:
        expanded_line = expand_line(line, var_dict)
        result += expanded_line + "\n"
    return result


def expand_env_vars(spec: MerlinSpec) -> MerlinSpec:
    """
    Expand environment variables in all sections of a spec.

    This function processes all sections of a given spec object and expands
    environment variables (e.g., `$HOME` or `~`) in string values. It skips
    expansion for values associated with the keys 'cmd' or 'restart', as these
    are typically shell scripts where environment variable expansion would
    already occur during execution.

    Args:
        spec (spec.specification.MerlinSpec): The spec object
            containing sections to process.

    Returns:
        (spec.specification.MerlinSpec): The updated spec object with environment variables expanded in all
            applicable sections.
    """

    def recurse(section):
        if section is None:
            return section
        if isinstance(section, str):
            return expandvars(expanduser(section))
        if isinstance(section, dict):
            for key, val in section.items():
                if key in ["cmd", "restart"]:
                    continue
                section[key] = recurse(val)
        elif isinstance(section, list):
            for i, elem in enumerate(deepcopy(section)):
                section[i] = recurse(elem)
        return section

    for name, section in spec.sections.items():
        setattr(spec, name, recurse(section))
    return spec


def determine_user_variables(*user_var_dicts: List[Dict]) -> Dict:
    """
    Determine user-defined variables from multiple dictionaries.

    This function takes an arbitrary number of dictionaries containing user-defined
    variables and resolves them in order, handling variable references and expansions
    (e.g., environment variables, user home directory shortcuts). Variable names are
    converted to uppercase, and reserved words cannot be reassigned.

    Args:
        user_var_dicts: One or more dictionaries of user variables. Each dictionary
            contains key-value pairs where the key is the variable name, and the value
            is the variable's definition.

    Returns:
        A dictionary of resolved user variables, with variable names in uppercase
            and all references expanded.

    Raises:
        ValueError: If a reserved word is attempted to be reassigned.

    Example:
        ```python
        >>> user_vars_1 = {'OUTPUT_PATH': './studies', 'N_SAMPLES': 10}
        >>> user_vars_2 = {'TARGET': 'target_dir', 'PATH': '$(SPECROOT)/$(TARGET)'}
        >>> determine_user_variables(user_vars_1, user_vars_2)
        {'OUTPUT_PATH': './studies', 'N_SAMPLES': '10',
         'TARGET': 'target_dir', 'PATH': '$(SPECROOT)/target_dir'}
        ```
    """
    all_var_dicts = dict(ChainMap(*user_var_dicts))
    determined_results = {}
    for key, val in all_var_dicts.items():
        if key in RESERVED:
            raise ValueError(f"Cannot reassign value of reserved word '{key}'! Reserved words are: {RESERVED}.")
        new_val = str(val)
        if contains_token(new_val):
            for determined_key, determined_val in determined_results.items():
                var_determined_key = var_ref(determined_key)
                if var_determined_key in new_val:
                    new_val = new_val.replace(var_determined_key, determined_val)
        new_val = expandvars(expanduser(new_val))
        determined_results[key.upper()] = new_val
    return determined_results


def parameter_substitutions_for_sample(
    sample: List[str], labels: List[str], sample_id: int, relative_path_to_sample: str
) -> List[Tuple[str, str]]:
    """
    Generate parameter substitutions for a specific sample.

    This function creates a list of substitution pairs for a given sample,
    mapping variable references (e.g., `$(LABEL)`) to their corresponding
    values in the sample. It also includes metadata substitutions such as
    the sample ID and the relative path to the sample.

    Args:
        sample: A list of values representing the sample.
        labels: A list of column labels corresponding to the sample values.
        sample_id: The unique integer ID of the sample.
        relative_path_to_sample: The relative path to the sample.

    Returns:
        A list of tuples, where each tuple contains a variable reference
            (e.g., `$(LABEL)`) and its corresponding value.

    Example:
        ```python
        >>> sample = [10, 20]
        >>> labels = ["X", "Y"]
        >>> sample_id = 0
        >>> relative_path_to_sample = "/0/3/4/8/9/"
        >>> parameter_substitutions_for_sample(sample, labels, sample_id, relative_path_to_sample)
        [
            ("$(X)", "10"),
            ("$(Y)", "20"),
            ("$(MERLIN_SAMPLE_ID)", "0"),
            ("$(MERLIN_SAMPLE_PATH)", "/0/3/4/8/9/")
        ]
        ```
    """
    substitutions = []
    for label, axis in zip(labels, sample):
        substitutions.append((f"$({label})", str(axis)))

    # The 0:N (integer) id of a sample
    substitutions.append(("$(MERLIN_SAMPLE_ID)", str(sample_id)))

    # Specific path to the sample, ie /0/3/4/8/9/
    substitutions.append(("$(MERLIN_SAMPLE_PATH)", relative_path_to_sample))

    return substitutions


def parameter_substitutions_for_cmd(glob_path: str, sample_paths: str) -> List[Tuple[str, str]]:
    """
    Generate parameter substitutions for a Merlin command.

    This function creates a list of substitution pairs for a Merlin command,
    mapping variable references to their corresponding values. It also includes
    predefined return codes for various Merlin states.

    Substitutions:
        - `$(MERLIN_GLOB_PATH)`: The provided `glob_path`.
        - `$(MERLIN_PATHS_ALL)`: The provided `sample_paths`.
        - `$(MERLIN_SUCCESS)`: The return code for a successful operation.
        - `$(MERLIN_RESTART)`: The return code for a restart operation.
        - `$(MERLIN_SOFT_FAIL)`: The return code for a soft failure.
        - `$(MERLIN_HARD_FAIL)`: The return code for a hard failure.
        - `$(MERLIN_RETRY)`: The return code for a retry operation.
        - `$(MERLIN_STOP_WORKERS)`: The return code for stopping workers.
        - `$(MERLIN_RAISE_ERROR)`: The return code for raising an error.

    Args:
        glob_path: A glob pattern that yields the paths to all Merlin samples.
        sample_paths: A delimited string containing the paths to all samples.

    Returns:
        A list of tuples, where each tuple contains a variable reference and its
            corresponding value.

    Example:
        ```python
        >>> glob_path = "/path/to/samples/*"
        >>> sample_paths = "/path/to/sample1:/path/to/sample2"
        >>> parameter_substitutions_for_cmd(glob_path, sample_paths)
        [
            ("$(MERLIN_GLOB_PATH)", "/path/to/samples/*"),
            ("$(MERLIN_PATHS_ALL)", "/path/to/sample1:/path/to/sample2"),
            ("$(MERLIN_SUCCESS)", "0"),
            ("$(MERLIN_RESTART)", "100"),
            ("$(MERLIN_SOFT_FAIL)", "101"),
            ("$(MERLIN_HARD_FAIL)", "102"),
            ("$(MERLIN_RETRY)", "104"),
            ("$(MERLIN_STOP_WORKERS)", "105"),
            ("$(MERLIN_RAISE_ERROR)", "106")
        ]
        ```
    """
    substitutions = []
    substitutions.append(("$(MERLIN_GLOB_PATH)", glob_path))
    substitutions.append(("$(MERLIN_PATHS_ALL)", sample_paths))
    # Return codes
    substitutions.append(("$(MERLIN_SUCCESS)", str(int(ReturnCode.OK))))
    substitutions.append(("$(MERLIN_RESTART)", str(int(ReturnCode.RESTART))))
    substitutions.append(("$(MERLIN_SOFT_FAIL)", str(int(ReturnCode.SOFT_FAIL))))
    substitutions.append(("$(MERLIN_HARD_FAIL)", str(int(ReturnCode.HARD_FAIL))))
    substitutions.append(("$(MERLIN_RETRY)", str(int(ReturnCode.RETRY))))
    substitutions.append(("$(MERLIN_STOP_WORKERS)", str(int(ReturnCode.STOP_WORKERS))))
    substitutions.append(("$(MERLIN_RAISE_ERROR)", str(int(ReturnCode.RAISE_ERROR))))
    return substitutions


# There's similar code inside study.py but the whole point of this function is to not use
# the MerlinStudy object so we disable this pylint error
# pylint: disable=duplicate-code
def expand_spec_no_study(filepath: str, override_vars: Dict[str, str] = None) -> str:
    """
    Get the expanded text of a spec without creating
    a MerlinStudy. Expansion is limited to user variables
    (the ones defined inside the yaml spec or at the command
    line).

    Expand a spec without creating a [`MerlinStudy`][study.study.MerlinStudy].

    This function processes a spec file to expand user-defined variables (those defined
    in the YAML spec or provided via `override_vars`) without creating a `MerlinStudy`
    object. It returns the expanded text of the specification.

    Args:
        filepath: The path to the YAML specification file.
        override_vars: A dictionary of variable overrides to apply during the expansion.
            These overrides replace or supplement the variables defined in the spec.

    Returns:
        The expanded YAML specification as a string, with user-defined variables resolved.
    """
    error_override_vars(override_vars, filepath)
    spec = MerlinSpec.load_specification(filepath)
    spec.environment = replace_override_vars(spec.environment, override_vars)
    spec_text = spec.dump()

    uvars = []
    if "variables" in spec.environment:
        uvars.append(spec.environment["variables"])
    if "labels" in spec.environment:
        uvars.append(spec.environment["labels"])
    evaluated_uvars = determine_user_variables(*uvars)

    return expand_by_line(spec_text, evaluated_uvars)


# pylint: enable=duplicate-code


def get_spec_with_expansion(filepath: str, override_vars: Dict[str, str] = None) -> MerlinSpec:
    """
    Load and expand a Merlin YAML specification with overrides, without creating a
    [`MerlinStudy`][study.study.MerlinStudy] object.

    This function returns a [`MerlinSpec`][spec.specification.MerlinSpec] object with
    variables expanded and overrides applied. It processes the YAML specification file
    and resolves user-defined variables without creating a `MerlinStudy` object.

    Args:
        filepath: The path to the YAML specification file.
        override_vars: A dictionary of variable overrides to apply during the expansion.
            These overrides replace or supplement the variables defined in the YAML spec.

    Returns:
        (spec.specification.MerlinSpec): A `MerlinSpec` object with expanded variables and applied overrides.
    """
    filepath = verify_filepath(filepath)
    expanded_spec_text = expand_spec_no_study(filepath, override_vars)
    return MerlinSpec.load_spec_from_string(expanded_spec_text)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
SQLite store implementations for Merlin entity models.

This module defines concrete `SQLiteStoreBase` subclasses for managing the
persistence of core entity models used in the Merlin application. Each store
is bound to a specific model and SQLite table, providing CRUD operations and
automated table creation.

See also:
    - merlin.backends.sqlite.sqlite_base_store: Base class
    - merlin.db_scripts.data_models: Data model definitions
"""

from merlin.backends.sqlite.sqlite_store_base import SQLiteStoreBase
from merlin.db_scripts.data_models import LogicalWorkerModel, PhysicalWorkerModel, RunModel, StudyModel


class SQLiteLogicalWorkerStore(SQLiteStoreBase[LogicalWorkerModel]):
    """
    A SQLite-based store for managing [`LogicalWorkerModel`][db_scripts.data_models.LogicalWorkerModel]
    objects.
    """

    def __init__(self):
        """Initialize the `SQLiteLogicalWorkerStore`."""
        super().__init__("logical_worker", LogicalWorkerModel)


class SQLitePhysicalWorkerStore(SQLiteStoreBase[PhysicalWorkerModel]):
    """
    A SQLite-based store for managing [`PhysicalWorkerModel`][db_scripts.data_models.PhysicalWorkerModel]
    objects.
    """

    def __init__(self):
        """Initialize the `SQLitePhysicalWorkerStore`."""
        super().__init__("physical_worker", PhysicalWorkerModel)


class SQLiteRunStore(SQLiteStoreBase[RunModel]):
    """
    A SQLite-based store for managing [`RunModel`][db_scripts.data_models.RunModel]
    objects.
    """

    def __init__(self):
        """Initialize the `SQLiteRunStore`."""
        super().__init__("run", RunModel)


class SQLiteStudyStore(SQLiteStoreBase[StudyModel]):
    """
    A SQLite-based store for managing [`StudyModel`][db_scripts.data_models.StudyModel]
    objects.
    """

    def __init__(self):
        """Initialize the `SQLiteStudyStore`."""
        super().__init__("study", StudyModel)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module provides utility functions and constants to manage broker configurations and connection strings
for various messaging systems, including RabbitMQ and Redis. It supports multiple connection protocols
and configurations, such as SSL, Unix sockets, and password inclusion.

The module defines constants for supported brokers and connection string templates, along with functions
to construct and retrieve connection strings and SSL configurations based on settings defined in the
`app.yaml` configuration file.
"""
from __future__ import print_function

import getpass
import logging
import os
import ssl
from os.path import expanduser
from typing import Dict, List, Optional, Union
from urllib.parse import quote

from merlin.config.configfile import CONFIG, get_ssl_entries


LOG: logging.Logger = logging.getLogger(__name__)

BROKERS: List[str] = ["rabbitmq", "redis", "rediss", "redis+socket", "amqps", "amqp"]

RABBITMQ_CONNECTION: str = "{conn}://{username}:{password}@{server}:{port}/{vhost}"
REDISSOCK_CONNECTION: str = "redis+socket://{path}?virtual_host={db_num}"
USER = getpass.getuser()


def read_file(filepath: str) -> str:
    """
    Safely reads the first line from a file and returns it with special characters URL-encoded.

    Args:
        filepath (str): The path to the file to be read.

    Returns:
        The first line of the file, stripped of leading and trailing whitespace,
            with special characters URL-encoded.
    """
    with open(filepath, "r") as f:  # pylint: disable=C0103
        line = f.readline().strip()
        return quote(line, safe="")


def get_rabbit_connection(include_password: bool, conn: str = "amqps") -> str:
    """
    Constructs and returns a RabbitMQ connection string based on broker configurations.

    This function reads broker configurations (such as server, port, username, password, and vhost)
    and formats them into a RabbitMQ connection string. Optionally, the password can be included
    in the connection string if `include_password` is set to `True`.

    Args:
        include_password (bool): Whether to include the password in the connection string.
        conn (str, optional): The connection protocol to use. Defaults to "amqps".
            Supported values are "amqp" and "amqps".

    Returns:
        A formatted RabbitMQ connection string.

    Raises:
        ValueError: If the password file path is not provided in the broker configuration, or if
            the password file does not exist or cannot be read.
    """
    LOG.debug(f"Broker: connection = {conn}")

    vhost = CONFIG.broker.vhost
    LOG.debug(f"Broker: vhost = {vhost}")

    username = CONFIG.broker.username
    LOG.debug(f"Broker: username = {username}")

    server = CONFIG.broker.server
    LOG.debug(f"Broker: server = {server}")

    try:
        password_filepath = CONFIG.broker.password
        LOG.debug("Broker: password file path has been configured.")
        password_filepath = os.path.abspath(expanduser(password_filepath))
    except (AttributeError, KeyError) as exc:
        raise ValueError("Broker: No password provided for RabbitMQ") from exc

    try:
        password = read_file(password_filepath)
    except IOError as exc:
        raise ValueError(f"Broker: RabbitMQ password file {password_filepath} does not exist") from exc

    try:
        port = CONFIG.broker.port
        LOG.debug(f"Broker: RabbitMQ port = {port}")
    except (AttributeError, KeyError):
        if conn == "amqp":
            port = 5672
        else:
            port = 5671
        LOG.debug(f"Broker: RabbitMQ using default port = {port}")

    # Test configurations.
    rabbitmq_config = {
        "conn": conn,
        "vhost": vhost,
        "username": username,
        "password": "******",
        "server": server,
        "port": port,
    }

    if include_password:
        rabbitmq_config["password"] = password

    return RABBITMQ_CONNECTION.format(**rabbitmq_config)


def get_redissock_connection() -> str:
    """
    Constructs and returns a Redis connection string using a Unix socket.

    This function retrieves broker configurations, such as the database number (`db_num`) and
    the Unix socket file path (`path`), and formats them into a Redis connection string.

    If the database number is not specified in the configuration, it defaults to `0`.

    Returns:
        A formatted Redis connection string using a Unix socket.
    """
    try:
        db_num = CONFIG.broker.db_num
    except (AttributeError, KeyError):
        db_num = 0
        LOG.debug(f"Broker: redis+socket using default db_num = {db_num}")

    redis_config = {"db_num": db_num, "path": CONFIG.broker.path}

    return REDISSOCK_CONNECTION.format(**redis_config)


# flake8 complains this function is too complex, we don't gain much nesting any of this as a separate function,
# however, cyclomatic complexity examination is off to get around this
def get_redis_connection(include_password: bool, use_ssl: bool = False) -> str:  # noqa C901
    """
    Constructs and returns a Redis connection string, optionally using SSL and including a password.

    This function retrieves broker configurations (such as server, port, username, password, and database number)
    and formats them into a Redis connection string. The connection can be configured to use SSL (`rediss` protocol)
    and optionally include the password in the connection string.

    Args:
        include_password (bool): Whether to include the password in the connection string.
        use_ssl (bool, optional): Whether to use the `rediss` protocol (SSL).

    Returns:
        A formatted Redis connection string.
    """
    server = CONFIG.broker.server
    LOG.debug(f"Broker: server = {server}")

    urlbase = "rediss" if use_ssl else "redis"

    try:
        port = CONFIG.broker.port
        LOG.debug(f"Broker: redis port = {port}")
    except (AttributeError, KeyError):
        port = 6379
        LOG.debug(f"Broker: redis using default port = {port}")

    try:
        db_num = CONFIG.broker.db_num
    except (AttributeError, KeyError):
        db_num = 0
        LOG.debug(f"Broker: redis using default db_num = {db_num}")

    try:
        username = CONFIG.broker.username
    except (AttributeError, KeyError):
        username = ""

    try:
        password_filepath = CONFIG.broker.password
        try:
            password = read_file(password_filepath)
        except IOError:
            password = CONFIG.broker.password
        if include_password:
            spass = f"{username}:{password}@"
        else:
            spass = f"{username}:******@"
    except (AttributeError, KeyError):
        spass = ""
        LOG.debug(f"Broker: redis using default password = {spass}")

    return f"{urlbase}://{spass}{server}:{port}/{db_num}"


def get_connection_string(include_password: bool = True) -> str:
    """
    Constructs and returns a connection string based on the broker configuration.

    This function retrieves the connection string from the `CONFIG.broker.url` if available.
    Otherwise, it determines the connection string based on the broker name specified in the
    configuration file (`app.yaml`). If the broker name is not supported, a `ValueError` is raised.

    Args:
        include_password (bool): Whether to include the password in the connection string.

    Returns:
        A formatted connection string based on the broker configuration.

    Raises:
        ValueError: If the broker name is not supported.
    """
    try:
        return CONFIG.broker.url
    except AttributeError:
        # The broker may not have a url
        pass

    try:
        broker = CONFIG.broker.name.lower()
    except AttributeError:
        broker = ""

    if broker not in BROKERS:
        raise ValueError(f"Error: {broker} is not a supported broker.")
    return _sort_valid_broker(broker, include_password)


def _sort_valid_broker(broker: str, include_password: bool) -> str:
    """
    Determines and returns the appropriate connection string for a given broker.

    This function selects the connection string generation method based on the broker type
    provided as input. Supported brokers include RabbitMQ (`amqp` or `amqps`), Redis (`redis`),
    Redis over SSL (`rediss`), and Redis over a socket (`redis+socket`).

    Args:
        broker (str): The name of the broker. Must be one of the supported broker types:
            `rabbitmq`, `amqps`, `amqp`, `redis+socket`, `redis`, or `rediss`.
        include_password (bool): Whether to include the password in the connection string.

    Returns:
        A formatted connection string for the specified broker.
    """
    if broker in ("rabbitmq", "amqps"):
        return get_rabbit_connection(include_password, conn="amqps")

    if broker == "amqp":
        return get_rabbit_connection(include_password, conn="amqp")

    if broker == "redis+socket":
        return get_redissock_connection()

    if broker == "redis":
        return get_redis_connection(include_password)

    # broker must be rediss
    return get_redis_connection(include_password, use_ssl=True)


def get_ssl_config() -> Union[bool, Dict[str, Union[str, ssl.VerifyMode]]]:
    """
    Retrieves the SSL configuration for the broker based on the settings in the `app.yaml` configuration file.

    This function determines whether SSL should be used for the broker connection and, if applicable,
    returns the SSL configuration details. If the broker does not require SSL or is unsupported,
    the function returns `False`.

    Returns:
        This returns either:\n
            - `False` if SSL is not required or the broker is unsupported.
            - A dictionary containing SSL configuration details if SSL is required.
              The dictionary may include keys such as certificate paths and verification modes.
    """
    broker: Union[bool, str] = ""
    try:
        broker = CONFIG.broker.url.split(":")[0]
    except AttributeError:
        # The broker may not have a url
        pass

    try:
        broker = CONFIG.broker.name.lower()
    except AttributeError:
        # The broker may not have a name
        pass

    if broker not in BROKERS:
        return False

    certs_path: Optional[str]
    try:
        certs_path = CONFIG.celery.certs
    except AttributeError:
        certs_path = None

    broker_ssl: Union[bool, Dict[str, Union[str, ssl.VerifyMode]]] = get_ssl_entries(
        "Broker", broker, CONFIG.broker, certs_path
    )

    if not broker_ssl:
        broker_ssl = True

    if broker in ("rabbitmq", "rediss", "amqps"):
        return broker_ssl

    return False



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
This module contains Celery task definitions.

The purpose of this module is to convert the Directed Acyclic Graph
([`DAG`][study.dag.DAG]) provided by Maestro into smaller tasks that
Celery can manage.
"""
from __future__ import absolute_import, unicode_literals

import json
import logging
import os
from typing import Any, Callable, Dict, List, Optional

# Need to disable an overwrite warning here since celery has an exception that we need that directly
# overwrites a python built-in exception
from celery import Signature, Task, chain, chord, group, shared_task, signature
from celery.exceptions import MaxRetriesExceededError
from celery.exceptions import OperationalError as CeleryOperationalError
from celery.exceptions import TimeoutError as CeleryTimeoutError
from celery.result import AsyncResult
from filelock import FileLock, Timeout
from kombu.exceptions import OperationalError as KombuOperationalError
from redis.exceptions import TimeoutError as RedisTimeoutError

from merlin.common.enums import ReturnCode
from merlin.common.sample_index import SampleIndex, uniform_directories
from merlin.common.sample_index_factory import create_hierarchy
from merlin.config.utils import Priority, get_priority
from merlin.db_scripts.merlin_db import MerlinDatabase
from merlin.exceptions import HardFailException, InvalidChainException, RestartException, RetryException
from merlin.router import stop_workers
from merlin.spec.expansion import parameter_substitutions_for_cmd, parameter_substitutions_for_sample
from merlin.study.dag import DAG
from merlin.study.status import read_status, status_conflict_handler
from merlin.study.step import Step
from merlin.study.study import MerlinStudy
from merlin.utils import dict_deep_merge


retry_exceptions = (
    # Python Built-in Exceptions
    IOError,
    OSError,
    AttributeError,
    TimeoutError,
    FileNotFoundError,
    # Celery Exceptions
    CeleryOperationalError,
    CeleryTimeoutError,
    # Kombu Exceptions
    KombuOperationalError,
    # Redis Exceptions
    RedisTimeoutError,
    # Merlin Exceptions
    RetryException,
    RestartException,
)

LOG = logging.getLogger(__name__)

STOP_COUNTDOWN = 60

# TODO: most of the pylint errors that are disabled in this file are the ones listed below.
# We should refactor this file so that we use more functions to solve all of these errors
# R0912: too many branches
# R0913: too many arguments
# R0914: too many local variables
# R0915: too many statements


@shared_task(  # noqa: C901
    bind=True,
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    priority=get_priority(Priority.HIGH),
)
def merlin_step(self: Task, *args: Any, **kwargs: Any) -> ReturnCode:  # noqa: C901 pylint: disable=R0912,R0915
    """
    Executes a Merlin step.

    This task executes a step in the Merlin workflow, handling various
    outcomes such as success, retries, and failures. It can also manage
    chaining to the next step in the workflow.

    Notes:
        - If the step has already been completed, it will be skipped.

    Args:
        self: The current task instance.
        *args: Positional arguments, one of which should be an instance
            of [`Step`][study.step.Step].
        **kwargs: Optional keyword arguments that include:\n
            - adapter_config (`Dict`): Configuration for the adapter,
              defaulting to `{'type': 'local'}`.
            - next_in_chain ([`Step`][study.step.Step]): The next step in
                the workflow chain, if applicable.\n
            Example kwargs dict where `merlin_step` will be added to the
            current chord with `next_in_chain` as an argument:\n
            ```
            {
                "adapter_config": {
                    'type': 'local'
                },
                "next_in_chain": <Step object>
            }
            ```

    Returns:
        (common.enums.ReturnCode): The result of the step
            execution, which can indicate success, various failure modes,
            or a request to retry.
    """
    step: Optional[Step] = None
    LOG.debug(f"args is {len(args)} long")

    arg: Any
    for arg in args:
        if isinstance(arg, Step):
            step = arg
        else:
            LOG.debug(f"discard argument {arg}, not of type Step.")

    config: Dict[str, str] = kwargs.pop("adapter_config", {"type": "local"})
    next_in_chain: Optional[Step] = kwargs.pop("next_in_chain", None)

    if step:
        self.max_retries = step.max_retries
        step_name: str = step.name()
        step_dir: str = step.get_workspace()
        LOG.debug(f"merlin_step: step_name '{step_name}' step_dir '{step_dir}'")
        finished_filename: str = os.path.join(step_dir, "MERLIN_FINISHED")

        # if we've already finished this task, skip it
        result: ReturnCode
        if os.path.exists(finished_filename):
            LOG.info(f"Skipping step '{step_name}' in '{step_dir}'.")
            result = ReturnCode.OK
        else:
            LOG.info(f"Executing step '{step_name}' in '{step_dir}'...")
            result = step.execute(config)
            step.mstep.mark_end(result)

        if result == ReturnCode.OK:
            LOG.info(f"Step '{step_name}' in '{step_dir}' finished successfully.")
            # touch a file indicating we're done with this step
            with open(finished_filename, "a"):
                pass
        elif result == ReturnCode.DRY_OK:
            LOG.info(f"Dry-ran step '{step_name}' in '{step_dir}'.")
        elif result == ReturnCode.RESTART:
            step.restart = True
            try:
                LOG.info(
                    f"Step '{step_name}' in '{step_dir}' is being restarted ({self.request.retries + 1}/{self.max_retries})..."
                )
                step.mstep.mark_restart()
                self.retry(countdown=step.retry_delay, priority=get_priority(Priority.RETRY))
            except MaxRetriesExceededError:
                LOG.warning(
                    f"""*** Step '{step_name}' in '{step_dir}' exited with a MERLIN_RESTART command,
                    but has already reached its retry limit ({self.max_retries}). Continuing with workflow."""
                )
                result = ReturnCode.SOFT_FAIL
                # Need to call mark_end again since we switched from RESTART to SOFT_FAIL
                step.mstep.mark_end(result, max_retries=True)
        elif result == ReturnCode.RETRY:
            step.restart = False
            try:
                LOG.info(
                    f"Step '{step_name}' in '{step_dir}' is being retried ({self.request.retries + 1}/{self.max_retries})..."
                )
                step.mstep.mark_restart()
                self.retry(countdown=step.retry_delay, priority=get_priority(Priority.RETRY))
            except MaxRetriesExceededError:
                LOG.warning(
                    f"""*** Step '{step_name}' in '{step_dir}' exited with a MERLIN_RETRY command,
                    but has already reached its retry limit ({self.max_retries}). Continuing with workflow."""
                )
                result = ReturnCode.SOFT_FAIL
                # Need to call mark_end again since we switched from RETRY to SOFT_FAIL
                step.mstep.mark_end(result, max_retries=True)
        elif result == ReturnCode.SOFT_FAIL:
            LOG.warning(f"*** Step '{step_name}' in '{step_dir}' soft failed. Continuing with workflow.")
        elif result == ReturnCode.HARD_FAIL:
            # stop all workers attached to this queue
            step_queue = step.get_task_queue()
            LOG.error(f"*** Step '{step_name}' in '{step_dir}' hard failed. Quitting workflow.")
            LOG.error(f"*** Shutting down all workers connected to this queue ({step_queue}) in {STOP_COUNTDOWN} secs!")
            shutdown = shutdown_workers.s([step_queue])
            shutdown.set(queue=step_queue)
            shutdown.apply_async(countdown=STOP_COUNTDOWN)
            raise HardFailException
        elif result == ReturnCode.STOP_WORKERS:
            LOG.warning(f"*** Shutting down all workers in {STOP_COUNTDOWN} secs!")
            shutdown = shutdown_workers.s(None)
            shutdown.set(queue=step.get_task_queue())
            shutdown.apply_async(countdown=STOP_COUNTDOWN)
        elif result == ReturnCode.RAISE_ERROR:
            LOG.warning("*** Raising an error ***")
            raise Exception("Exception raised by request from the user")
        else:
            LOG.warning(f"**** Step '{step_name}' in '{step_dir}' had unhandled exit code {result}. Continuing with workflow.")

        # queue off the next task in a chain while adding it to the current chord so that the chordfinisher actually
        # waits for the next task in the chain
        if next_in_chain is not None:
            if self.request.is_eager:
                LOG.debug(f"calling next_in_chain {signature(next_in_chain)}")
                next_in_chain.delay()
            else:
                LOG.debug(f"adding {next_in_chain} to chord")
                self.add_to_chord(next_in_chain, lazy=False)
        return result

    LOG.error("Failed to find step!")
    return None


def is_chain_expandable(chain_: List[Step], labels: List[str]) -> bool:
    """
    Determine if the steps in the given chain are expandable.

    A chain is considered expandable if all steps within the chain require
    expansion. Conversely, if none of the steps require expansion, the chain
    is not expandable. If there is a mix of steps that require expansion and
    those that do not, an `InvalidChainException` is raised, indicating that
    the chain is incompatible.

    Args:
        chain_ (List[study.step.Step]): A list of [`Step`][study.step.Step]
            objects representing a chain of dependent steps.
        labels: The labels associated with the steps in the chain, used to
            determine if expansion is needed.

    Returns:
        True if all steps in the chain are expandable, False if none are
            expandable.

    Raises:
        InvalidChainException: If there is a mix of steps that require
            expansion and those that do not, indicating an incompatible chain.
    """

    array_of_bools = [step.check_if_expansion_needed(labels) for step in chain_]

    needs_expansion = all(array_of_bools)

    if needs_expansion is False:
        # if we're not expanding, but at least one step needed expansion, then
        # this is an incompatible chain
        incompatible_chain = any(array_of_bools)

        if incompatible_chain is True:
            LOG.error(
                "INCOMPATIBLE CHAIN - all tasks in a chain need to either "
                "be merlin expanded or all need to not be merlin expanded. "
                "Please report this to merlin@llnl.gov"
            )
            raise InvalidChainException

    return needs_expansion


def prepare_chain_workspace(sample_index: SampleIndex, chain_: List[Step]):
    """
    Prepares a user's workspace for each step in the given chain of dependent steps.

    This function iterates through a list of [`Step`][study.step.Step] objects and
    prepares the necessary workspace for each step by creating directories and writing
    sample index files.

    Args:
        sample_index (common.sample_index.SampleIndex): An object that manages sample
            indexing and workspace preparation.
        chain_ (List[study.step.Step]): A list of [`Step`][study.step.Step] objects
            representing a chain of dependent steps. Each step's workspace will be prepared.
    """
    # TODO: figure out faster way to create these directories (probably using
    # yet another task)
    for step in chain_:
        workspace = step.get_workspace()
        LOG.debug(f"Preparing workspace in {workspace}...")

        # If we need to expand it, initialize the workspace for the samples
        sample_index.name = workspace
        sample_index.write_directories()
        sample_index.write_multiple_sample_index_files()
        LOG.debug(f"...workspace {workspace} prepared.")


@shared_task(
    bind=True,
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    priority=get_priority(Priority.LOW),
)
def add_merlin_expanded_chain_to_chord(  # pylint: disable=R0913,R0914
    self: Task,
    task_type: Signature,
    chain_: List[Step],
    samples: List[Any],
    labels: List[str],
    sample_index: SampleIndex,
    adapter_config: Dict,
    min_sample_id: int,
):
    """
    Expand tasks in a chain and add the expanded tasks to the current chord.

    This Celery task recursively expands a chain of tasks based on provided
    sample values and their corresponding labels. The expanded tasks are
    configured with specific parameters and added to the current chord for
    execution. The function handles both the expansion of tasks and the
    management of task dependencies.

    Args:
        self: The current task instance.
        task_type: The Celery task signature type for the new tasks to be
            created.
        chain_ (List[study.step.Step]): A list of tasks to expand into a chain.
        samples: The sample values to use for each new task.
        labels: The sample labels corresponding to the samples.
        sample_index (common.sample_index.SampleIndex): The sample index that
            contains the directory structure for tasks.
        adapter_config: Configuration settings for the adapter used in task
            execution.
        min_sample_id: An offset to use for the sample index.
    """
    num_samples = len(samples)
    # Use the index to get a path to each sample
    LOG.debug(f"recursing with {num_samples} samples {samples}")
    if sample_index.is_grandparent_of_leaf or sample_index.is_parent_of_leaf:
        all_chains = []
        LOG.debug(f"gathering up {num_samples} relative paths")
        relative_paths = [
            os.path.dirname(sample_index.get_path_to_sample(sample_id + min_sample_id)) for sample_id in range(num_samples)
        ]
        top_lvl_workspace = chain_[0].get_workspace()
        LOG.debug(f"recursing grandparent with relative paths {relative_paths}")
        for step in chain_:
            # Make a list of new task objects with modified cmd and workspace
            # based off of the parameter substitutions and relative_path for
            # a given sample.
            workspace = step.get_workspace()
            LOG.debug(f"expanding step {step.name()} in workspace {workspace}")
            new_chain = []
            for sample_id, sample in enumerate(samples):
                new_step = task_type.s(
                    step.clone_changing_workspace_and_cmd(
                        new_workspace=os.path.join(workspace, relative_paths[sample_id]),
                        cmd_replacement_pairs=parameter_substitutions_for_sample(
                            sample,
                            labels,
                            sample_id + min_sample_id,
                            relative_paths[sample_id],
                        ),
                    ),
                    adapter_config=adapter_config,
                    top_lvl_workspace=top_lvl_workspace,
                )
                new_step.set(queue=step.get_task_queue())
                new_step.set(task_id=os.path.join(workspace, relative_paths[sample_id]))
                new_chain.append(new_step)

            all_chains.append(new_chain)

        # Only need to condense status files if there's more than 1 sample
        if num_samples > 1:
            condense_sig = condense_status_files.s(
                sample_index=sample_index,
                workspace=top_lvl_workspace,
                condensed_workspace=chain_[0].mstep.condensed_workspace,
            ).set(
                queue=chain_[0].get_task_queue(),
            )
        else:
            condense_sig = None

        LOG.debug("adding chain to chord")
        chain_1d = get_1d_chain(all_chains)
        launch_chain(self, chain_1d, condense_sig=condense_sig)
        LOG.debug("chain added to chord")
    else:
        # recurse down the sample_index hierarchy
        try:
            LOG.debug("recursing down sample_index hierarchy")
            for next_index in sample_index.children.values():
                next_index_name_before = next_index.name
                next_index.name = os.path.join(sample_index.name, next_index.name)
                LOG.debug("generating next step")
                next_step = add_merlin_expanded_chain_to_chord.s(
                    task_type,
                    chain_,
                    samples[next_index.min - min_sample_id : next_index.max - min_sample_id],
                    labels,
                    next_index,
                    adapter_config,
                    next_index.min,
                )
                next_step.set(queue=chain_[0].get_task_queue())
                LOG.debug(f"recursing with range {next_index.min}:{next_index.max}, {next_index.name} {signature(next_step)}")
                LOG.debug(f"queuing samples[{next_index.min}:{next_index.max}] in for {chain_} in {next_index.name}...")
                if self.request.is_eager:
                    next_step.delay()
                else:
                    self.add_to_chord(next_step, lazy=False)
                LOG.debug(f"queued for samples[{next_index.min}:{next_index.max}] in for {chain_} in {next_index.name}")
        except retry_exceptions as e:
            # Reset the index to what it was before so we don't accidentally create a bunch of extra samples upon restart
            next_index.name = next_index_name_before
            raise e

    return ReturnCode.OK


def add_simple_chain_to_chord(self: Task, task_type: Signature, chain_: List[Step], adapter_config: Dict):
    """
    Add a chain of tasks to the current chord for execution.

    This function takes a list of tasks, modifies their signatures based on
    provided parameters, and adds them to the current chord. Each task in the
    chain is transformed into a new task signature with specific configurations
    such as queue and task ID.

    This function takes a list of steps and creates signatures based on the
    parameters they provide, such as queue and workspace. It then adds these
    signatures to the current chord for later execution.

    Args:
        self: The current task instance invoking this method.
        task_type: The Celery task signature type that the new tasks should be
            based on.
        chain_ (List[study.step.Step]): A list of tasks to expand into a chain.
            Each task should provide necessary parameters for signature creation.
        adapter_config: Configuration settings for the adapter used in task
            execution.
    """
    LOG.debug(f"simple chain with {chain_}")
    all_chains = []
    for step in chain_:
        # Make a list of new task signatures with modified cmd and workspace
        # based off of the parameter substitutions and relative_path for
        # a given sample.

        new_steps = [
            task_type.s(step, adapter_config=adapter_config).set(
                queue=step.get_task_queue(),
                task_id=step.get_workspace(),
            )
        ]
        all_chains.append(new_steps)
    chain_1d = get_1d_chain(all_chains)
    launch_chain(self, chain_1d)


def launch_chain(self: Task, chain_1d: List[Signature], condense_sig: Signature = None):
    """
    Launch a 1D chain of task signatures appropriately based on the execution context.

    This function handles the launching of a list of task signatures in a
    one-dimensional chain. The behavior varies depending on whether the
    execution is local or remote, and whether the tasks involve sample
    processing that requires condensing status files.

    Args:
        self: The current task instance invoking this method.
        chain_1d: A one-dimensional list of task signatures to be launched.
        condense_sig: A signature for condensing the status files after task execution.
            If None, condensing is not required.
    """
    # If there's nothing in the chain then we won't have to launch anything so check that first
    if chain_1d:
        # Case 1: local run; launch signatures instantly
        if self.request.is_eager:
            for sig in chain_1d:
                sig.delay()
        # Case 2: non-local run; signatures need to be added to the current chord
        else:
            # Case a: we're dealing with a sample hierarchy and need to condense status files when we're done executing tasks
            if condense_sig:
                # This chord makes it so we'll process all tasks in chain_1d, then condense the status files when they're done
                sample_chord = chord(chain_1d, condense_sig)
                self.add_to_chord(sample_chord, lazy=False)
            # Case b: no condensing is needed so just add all the signatures to the chord
            else:
                for sig in chain_1d:
                    self.add_to_chord(sig, lazy=False)


def get_1d_chain(all_chains: List[List[Signature]]) -> List[Signature]:
    """
    Convert a 2D list of task chains into a 1D list of task signatures.

    This function takes a two-dimensional list of task signatures, where each
    inner list represents a parallel group of tasks. It transforms this structure
    into a one-dimensional list suitable for creating a linear chain of tasks.
    If there is only one chain, it returns that chain directly. If there are
    multiple chains, it sets up dependencies between tasks to ensure proper
    execution order.

    Notes:
        - The function processes the chains in reverse order to correctly
          set up the dependencies before adding them to the final list.

    Args:
        all_chains: A two-dimensional list of task signatures, where each inner
            list represents a group of tasks that can be executed in parallel.

    Returns:
        A one-dimensional list of task signatures representing a chain of tasks,
            with dependencies set up for proper execution order.
    """
    chain_steps = []
    if len(all_chains) == 1:
        # Steps will be enqueued in a single parallel group
        chain_steps = all_chains[0]

    if len(all_chains) > 1:
        # in this case, we need to make a chain.
        # Celery chains do not natively update any chords they have to be in,
        # which can lead to a chord finishing prematurely,
        # causing future steps to execute before their dependencies are resolved.
        # Celery does provide an API to add a method to a chord dynamically
        # during execution of a task belonging to that chord,
        # so we set up a chain by passing the child member of a chain in as an
        # argument to the signature of the parent member of a chain.
        length = len(all_chains[0])
        for i in range(length):
            # Do the following in reverse order because the replace method
            # generates a new task signature, so we need to make
            # sure we are modifying task signatures before adding them to the
            # kwargs.
            for j in reversed(range(len(all_chains))):
                if j < len(all_chains) - 1:
                    # fmt: off
                    new_kwargs = signature(all_chains[j][i]).kwargs.update(
                        {"next_in_chain": all_chains[j + 1][i]}
                    )
                    # fmt: on
                    all_chains[j][i] = all_chains[j][i].replace(kwargs=new_kwargs)
            chain_steps.append(all_chains[0][i])

    return chain_steps


def gather_statuses(sample_index: SampleIndex, workspace: str, condensed_workspace: str, files_to_remove: List[str]) -> Dict:
    """
    Traverse the sample index and gather all statuses into a single dictionary.

    This function iterates through the provided
    [`SampleIndex`][common.sample_index.SampleIndex] object,
    reading status files from each sample's workspace. It condenses
    the statuses into a single dictionary while tracking which files
    need to be removed after condensing. The function ensures that
    only completed statuses are included in the condensed output.

    Args:
        sample_index (common.sample_index.SampleIndex): A
            [`SampleIndex`][common.sample_index.SampleIndex] object
            representing the specific sample hierarchy to traverse.
        workspace: The full path to the workspace for the step being
            condensed.
        condensed_workspace: A shortened version of the workspace
            path that will be used in the status files.
        files_to_remove: A list that will be populated with file paths
            of status files that need to be removed after condensing.

    Returns:
        A dictionary containing the condensed statuses gathered
            from the status files.

    Raises:
        TimeoutError: If a timeout occurs while reading a status file,
            triggering a restart of the task.
        FileNotFoundError: If a status file is not found during the
            condensing process.
    """
    LOG.info(f"Gathering statuses to condense for '{condensed_workspace}'")
    condensed_statuses = {}
    for path, _ in sample_index.traverse(conditional=lambda c: c.is_parent_of_leaf):
        # Read in the status data
        sample_workspace = f"{workspace}/{path}"
        status_filepath = f"{sample_workspace}/MERLIN_STATUS.json"
        lock_filepath = f"{sample_workspace}/status.lock"
        if os.path.exists(status_filepath):
            try:
                # NOTE: instead of leaving statuses as dicts read in by JSON, maybe they should each be their own object
                status = read_status(status_filepath, lock_filepath, raise_errors=True)

                # This for loop is just to get the step name that we don't have; it's really not even looping
                for step_name in status:
                    try:
                        # Make sure the status for this sample workspace is in a finished state (not initialized or running)
                        if status[step_name][f"{condensed_workspace}/{path}"]["status"] not in ("INITIALIZED", "RUNNING"):
                            # Add the status data to the statuses we'll write to the condensed file and remove this status file
                            dict_deep_merge(condensed_statuses, status, conflict_handler=status_conflict_handler)
                            files_to_remove.append(status_filepath)
                            files_to_remove.append(lock_filepath)  # Remove the lock files as well as the status files
                    except KeyError:
                        LOG.warning(f"Key error when reading from {sample_workspace}")
            except Timeout:
                # Raising this celery timeout instead will trigger a restart for this task
                raise TimeoutError  # pylint: disable=W0707
            except FileNotFoundError:
                LOG.warning(f"Could not find {status_filepath} while trying to condense. Restarting this task...")
                raise FileNotFoundError  # pylint: disable=W0707
        else:
            # Might be missing a status file in the output if we hit this but we don't want that
            # to fully crash the workflow
            LOG.debug(f"Could not find {status_filepath}, skipping this status file.")

    return condensed_statuses


@shared_task(
    bind=True,
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    priority=get_priority(Priority.LOW),
)
def condense_status_files(self: Task, *args: Any, **kwargs: Any) -> ReturnCode:  # pylint: disable=R0914,W0613
    """
    Condenses status files after a section of the sample tree has completed processing.

    This task gathers status information from a specified
    [`SampleIndex`][common.sample_index.SampleIndex] and condenses it into a single
    JSON file. It handles potential race conditions by using a file lock during
    the write operation. If the condensed status file already exists, it merges
    the new statuses with the existing ones.

    Notes:
        - The task will remove the original status files after condensing them
          into the JSON file.

    Args:
        self: The current task instance.
        *args: Additional positional arguments (not used in this task).
        **kwargs: Keyword arguments containing:\n
            - `sample_index` ([`SampleIndex`][common.sample_index.SampleIndex]):
                The [`SampleIndex`][common.sample_index.SampleIndex] object used
                for gathering statuses.
            - `workspace` (str): The workspace path for the step.
            - `condensed_workspace` (str): The workspace path for the
              condensed status.

    Returns:
        (common.enums.ReturnCode): A [`ReturnCode.OK`][common.enums.ReturnCode]
            message if the operation was successful. None, otherwise.

    Raises:
        TimeoutError: If the file lock cannot be acquired within the
            specified timeout period, which triggers a task restart.
    """
    # Get the sample index object that we'll use for condensing
    sample_index = kwargs.pop("sample_index", None)
    if not sample_index:
        LOG.warning("Sample index not found. Cannot condense status files.")
        return None

    # Get the full step (or step/parameter) workspace
    workspace = kwargs.pop("workspace", None)
    if not workspace:
        LOG.warning("Workspace not found. Cannot condense status files.")
        return None

    # Get a condensed version of the workspace
    condensed_workspace = kwargs.pop("condensed_workspace", None)
    if not condensed_workspace:
        LOG.warning("Condensed workspace not provided. Cannot condense status files.")
        return None

    # Read in all the statuses from this sample index
    files_to_remove = []
    condensed_statuses = gather_statuses(sample_index, workspace, condensed_workspace, files_to_remove)

    # If there are statuses to write to the condensed status file then write them
    if condensed_statuses:
        condensed_status_filepath = f"{workspace}/MERLIN_STATUS.json"
        condensed_lock_file = f"{workspace}/status.lock"
        lock = FileLock(condensed_lock_file)  # pylint: disable=E0110
        try:
            # Lock the file to avoid race conditions
            with lock.acquire(timeout=20):
                # If the condensed file already exists, grab the statuses from it
                LOG.info(f"Condensing statuses for '{condensed_workspace}' to '{condensed_status_filepath}'")
                if os.path.exists(condensed_status_filepath):
                    with open(condensed_status_filepath, "r") as condensed_status_file:
                        existing_condensed_statuses = json.load(condensed_status_file)
                    # Merging the statuses we're condensing into the already existing statuses
                    # because it's faster at scale than vice versa
                    dict_deep_merge(existing_condensed_statuses, condensed_statuses, conflict_handler=status_conflict_handler)
                    condensed_statuses = existing_condensed_statuses

                # Write the condensed statuses to the condensed status file
                with open(condensed_status_filepath, "w") as condensed_status_file:
                    json.dump(condensed_statuses, condensed_status_file)

                # Remove the status files we just condensed
                for file_to_remove in files_to_remove:
                    LOG.debug(f"Removing '{file_to_remove}'.")
                    os.remove(file_to_remove)
        except Timeout:
            # Raising this celery timeout instead will trigger a restart for this task
            raise TimeoutError  # pylint: disable=W0707

    return ReturnCode.OK


@shared_task(
    bind=True,
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    priority=get_priority(Priority.LOW),
)
def expand_tasks_with_samples(  # pylint: disable=R0913,R0914
    self: Task,
    dag: DAG,
    chain_: List[str],
    samples: List[List[str]],
    labels: List[str],
    task_type: Callable,
    adapter_config: Dict,
    level_max_dirs: int,
):
    """
    Expands a chain of task names into a group of Celery chains, using samples
    and labels for variable substitution.

    This task determines whether the provided chain of tasks requires
    expansion based on the structure of the Directed Acyclic Graph ([`DAG`][study.dag.DAG]),
    samples, and labels. If expansion is needed, it generates and queues new tasks
    for each range of samples. Otherwise, it queues a simple chain task.

    Args:
        self: The current task instance.
        dag (study.dag.DAG): A Merlin Directed Acyclic Graph
            ([`DAG`][study.dag.DAG]) representing the workflow.
        chain_: A list of task names to be expanded into a
            Celery group of chains.
        samples: A list of lists containing Merlin sample values for
            variable substitution.
        labels: A list of strings representing the labels associated
            with each column in the samples.
        task_type: The Celery task type to create, currently expected
            to be [`merlin_step`][common.tasks.merlin_step].
        adapter_config: A configuration dictionary for Maestro
            script adapters.
        level_max_dirs: The maximum number of directories allowed per
            level in the sample hierarchy.
    """
    LOG.debug(f"expand_tasks_with_samples called with chain,{chain_}\n")
    # Figure out how many directories there are, make a glob string
    directory_sizes = uniform_directories(len(samples), bundle_size=1, level_max_dirs=level_max_dirs)

    glob_path = "*/" * len(directory_sizes)

    LOG.debug("creating sample_index")
    # Write a hierarchy to get the all paths string
    sample_index = create_hierarchy(
        len(samples),
        bundle_size=1,
        directory_sizes=directory_sizes,
        root="",
        n_digits=len(str(level_max_dirs)),
    )

    LOG.debug("creating sample_paths")
    sample_paths = sample_index.make_directory_string()

    LOG.debug("assembling steps")
    # the steps in the chain
    steps = [dag.step(name) for name in chain_]

    # sub in globs prior to expansion
    # sub the glob command
    steps = [
        step.clone_changing_workspace_and_cmd(cmd_replacement_pairs=parameter_substitutions_for_cmd(glob_path, sample_paths))
        for step in steps
    ]

    # workspaces = [step.get_workspace() for step in steps]
    # LOG.debug(f"workspaces : {workspaces}")

    needs_expansion = is_chain_expandable(steps, labels)

    LOG.debug(f"needs_expansion {needs_expansion}")

    if needs_expansion:
        # prepare_chain_workspace(sample_index, steps)
        sample_index.name = ""
        LOG.debug("queuing merlin expansion tasks")
        found_tasks = False
        conditions = [
            lambda c: c.is_great_grandparent_of_leaf,
            lambda c: c.is_grandparent_of_leaf,
            lambda c: c.is_parent_of_leaf,
            lambda c: c.is_leaf,
        ]
        for condition in conditions:
            if not found_tasks:
                for next_index_path, next_index in sample_index.traverse(conditional=condition):
                    LOG.info(
                        f"generating next step for range {next_index.min}:{next_index.max} {next_index.max - next_index.min}"
                    )
                    next_index.name = next_index_path

                    sig = add_merlin_expanded_chain_to_chord.s(
                        task_type,
                        steps,
                        samples[next_index.min : next_index.max],
                        labels,
                        next_index,
                        adapter_config,
                        next_index.min,
                    )
                    sig.set(queue=steps[0].get_task_queue())

                    if self.request.is_eager:
                        sig.delay()
                    else:
                        LOG.info(f"queuing expansion task {next_index.min}:{next_index.max}")
                        self.add_to_chord(sig, lazy=False)
                    LOG.info(f"merlin expansion task {next_index.min}:{next_index.max} queued")
                    found_tasks = True
    else:
        LOG.debug("queuing simple chain task")
        add_simple_chain_to_chord(self, task_type, steps, adapter_config)
        LOG.debug("simple chain task queued")


# Pylint complains that "self" is unused but it's needed behind the scenes with celery
@shared_task(
    bind=True,
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    acks_late=False,
    reject_on_worker_lost=False,
    name="merlin:shutdown_workers",
    priority=get_priority(Priority.HIGH),
)
def shutdown_workers(self: Task, shutdown_queues: List[str]):  # pylint: disable=W0613
    """
    Initiates the shutdown of Celery workers.

    This task wraps the [`stop_celery_workers`][study.celeryadapter.stop_celery_workers]
    function, allowing for the graceful shutdown of specified Celery worker queues. It is
    acknowledged immediately upon execution, ensuring that it will not be requeued, even
    if executed by a worker.

    Args:
        self: The current task instance.
        shutdown_queues: A list of specific queues to shut down. If None, all queues will
            be shut down.
    """
    if shutdown_queues is not None:
        LOG.warning(f"Shutting down workers in queues {shutdown_queues}!")
    else:
        LOG.warning("Shutting down workers in all queues!")
    return stop_workers("celery", None, shutdown_queues, None)


# Pylint complains that these args are unused but celery passes args
# here behind the scenes and won't work if these aren't here
@shared_task(
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    name="merlin:chordfinisher",
    priority=get_priority(Priority.LOW),
)
def chordfinisher(*args: List, **kwargs: Dict) -> str:  # pylint: disable=W0613
    """
    Synchronization callback for Celery chords.

    This function serves as a synchronization point between groups of tasks
    in a Celery workflow. In Celery, using `chain(group, group)` does not
    guarantee that the second group will execute only after the first group
    has completed. Instead, both groups are executed independently.

    To enforce a synchronization point between these groups, this function
    is used as a callback in a chord. It allows for the declaration of chains
    of groups dynamically, ensuring that subsequent tasks wait for the
    completion of all tasks in the preceding groups.

    Args:
        *args: Variable length argument list. Needed by Celery.
        **kwargs: Arbitrary keyword arguments. Needed by Celery.

    Returns:
        A constant string "SYNC" indicating the synchronization point
            has been reached.
    """
    return "SYNC"


@shared_task(
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    name="merlin:mark_run_as_complete",
    priority=get_priority(Priority.LOW),
)
def mark_run_as_complete(study_workspace: str) -> str:
    """
    Mark this run as complete and save that to the database.

    Args:
        study_workspace: The output workspace for this run.

    Returns:
        A string denoting that this run has completed.
    """
    merlin_db = MerlinDatabase()
    run_entity = merlin_db.get("run", study_workspace)
    run_entity.run_complete = True
    run_entity.save()
    return "Run Completed"


@shared_task(
    autoretry_for=retry_exceptions,
    retry_backoff=True,
    name="merlin:queue_merlin_study",
    priority=get_priority(Priority.LOW),
)
def queue_merlin_study(study: MerlinStudy, adapter: Dict) -> AsyncResult:
    """
    Launch a chain of tasks based on a MerlinStudy.

    This Celery task initiates a series of tasks derived from a
    [`MerlinStudy`][study.study.MerlinStudy] object. It processes
    the study's Directed Acyclic Graph ([`DAG`][study.dag.DAG])
    to group tasks and convert them into a chain of Celery tasks
    for execution.

    Args:
        study: The study object containing samples, sample labels,
            and the Directed Acyclic Graph ([`DAG`][study.dag.DAG])
            structure that defines the task dependencies.
        adapter: An adapter object used to facilitate interactions with
            the study's data or processing logic.

    Returns:
        An instance representing the asynchronous result of the task chain,
            allowing for tracking and management of the task's execution.
    """
    samples = study.samples
    sample_labels = study.sample_labels
    egraph = study.dag
    LOG.info("Calculating task groupings from DAG.")
    groups_of_chains = egraph.group_tasks("_source")

    # magic to turn graph into celery tasks
    LOG.info("Converting graph to tasks.")
    celery_dag = chain(
        chord(
            group(
                [
                    expand_tasks_with_samples.si(
                        egraph,
                        gchain,
                        samples,
                        sample_labels,
                        merlin_step,
                        adapter,
                        study.level_max_dirs,
                    ).set(queue=egraph.step(chain_group[0][0]).get_task_queue())
                    for gchain in chain_group
                ]
            ),
            chordfinisher.s().set(queue=egraph.step(chain_group[0][0]).get_task_queue()),
        )
        for chain_group in groups_of_chains[1:]
    )

    # Append the final task that marks the run as complete
    final_task = mark_run_as_complete.si(study.workspace).set(
        queue=egraph.step(
            groups_of_chains[-1][-1][-1]  # Use the task queue from the final step to execute this task
        ).get_task_queue()
    )
    celery_dag = celery_dag | final_task

    LOG.info("Launching tasks.")
    return celery_dag.delay(None)



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
SQLite connection context manager for the Merlin application.

This module defines the `SQLiteConnection` class, which provides a safe and reusable way to
establish and manage SQLite connections using a context manager. It ensures proper configuration
(e.g., enabling WAL mode and foreign key support), handles compatibility with Python versions,
and guarantees cleanup by closing the connection on exit.
"""

import logging
import sqlite3
import sys
from pathlib import Path
from types import TracebackType
from typing import Type


LOG = logging.getLogger(__name__)


class SQLiteConnection:
    """
    Context manager for establishing and safely closing a SQLite database connection.

    This class ensures SQLite connections are created with proper configuration, including:
    - WAL mode for better concurrency
    - Foreign key constraint enforcement
    - Dictionary-style row access via `sqlite3.Row`
    - Compatibility with Python versions < 3.12 and ≥ 3.12 regarding autocommit

    Attributes:
        conn (sqlite3.Connection): The active SQLite connection used within the context.

    Methods:
        __enter__:
            Opens and configures the SQLite connection when entering the context.

        __exit__:
            Closes the SQLite connection when exiting the context, handling any exceptions.
    """

    def __init__(self):
        """Initialize the SQLiteConnection context manager."""
        self.conn: sqlite3.Connection = None

    def __enter__(self) -> sqlite3.Connection:
        """
        Enters the runtime context related to this object and creates a sqlite connection.

        Returns:
            A sqlite connection.
        """
        from merlin.config.results_backend import get_connection_string  # pylint: disable=import-outside-toplevel

        # Get the SQLite database path and ensure the path exists
        db_path = get_connection_string()
        Path(db_path).parent.mkdir(parents=True, exist_ok=True)

        connection_kwargs = {"check_same_thread": False}
        if sys.version_info < (3, 12):  # Autocommit wasn't added until python 3.12
            connection_kwargs["isolation_level"] = None
        else:
            connection_kwargs["autocommit"] = True

        # Create SQLite connection with optimized settings
        self.conn = sqlite3.connect(db_path, **connection_kwargs)

        # Enable WAL mode for better concurrent access
        self.conn.execute("PRAGMA journal_mode=WAL")
        # Enable foreign key constraints
        self.conn.execute("PRAGMA foreign_keys=ON")

        # This enables name-based access to columns
        self.conn.row_factory = sqlite3.Row

        return self.conn

    def __exit__(self, exc_type: Type[Exception], exc_value: Exception, traceback: TracebackType):
        """
        Exits the runtime context and performs cleanup.

        This method closes the connection if it's still open.

        Args:
            exc_type: The exception type raised, if any.
            exc_value: The exception instance raised, if any.
            traceback: The traceback object, if an exception was raised.
        """
        if self.conn:
            self.conn.close()



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
The `spec` package defines the structure, defaults, and functionality for working with
Merlin specification files.

Modules:
    all_keys.py: Defines all the valid keys for each block in a Merlin specification file,
        ensuring consistency and validation.
    defaults.py: Provides the default values for each block in a spec file, enabling workflows
        to execute even when fields are omitted.
    expansion.py: Handles the expansion of variables within a spec file, including user-defined,
        environment, and reserved variables, as well as parameter substitutions.
    override.py: Supports overriding variables in a spec file via the command-line interface,
        with functions for validation and replacement.
    specification.py: Contains the `MerlinSpec` class, which represents the raw data from a
        Merlin specification file and provides methods for interacting with it.
"""



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing database entities related to physical workers.

This module defines the `PhysicalWorkerManager` class, which provides high-level operations for
creating, retrieving, and deleting physical worker entities stored in the database. It acts as a
controller that encapsulates logic around
[`PhysicalWorkerEntity`][db_scripts.entities.physical_worker_entity.PhysicalWorkerEntity]
objects and their corresponding [`PhysicalWorkerModel`][db_scripts.data_models.PhysicalWorkerModel]
representations.

The manager interacts with the results backend and optionally references the main database
object to support operations that involve other entities, such as cleanup of related logical workers.
"""

from __future__ import annotations

import logging
from typing import Any, List

from merlin.db_scripts.data_models import PhysicalWorkerModel
from merlin.db_scripts.entities.physical_worker_entity import PhysicalWorkerEntity
from merlin.db_scripts.entity_managers.entity_manager import EntityManager
from merlin.exceptions import WorkerNotFoundError


LOG = logging.getLogger("merlin")

# Purposefully ignoring this pylint message as each entity will have different parameter requirements
# pylint: disable=arguments-differ,arguments-renamed


class PhysicalWorkerManager(EntityManager[PhysicalWorkerEntity, PhysicalWorkerModel]):
    """
    Manager for physical worker entities.

    This manager handles the creation, retrieval, and deletion of physical worker entities.
    It abstracts lower-level backend interactions and optionally performs cleanup logic
    that involves related logical workers through a reference to the main
    [`MerlinDatabase`][db_scripts.merlin_db.MerlinDatabase].

    Attributes:
        backend (backends.results_backend.ResultsBackend): The backend used for database operations.
        db (db_scripts.merlin_db.MerlinDatabase): Optional reference to the main database for cross-entity logic.

    Methods:
        create: Create a new physical worker if it does not already exist.
        get: Retrieve a physical worker entity by its ID or name.
        get_all: Retrieve all physical worker entities from the database.
        delete: Delete a specific physical worker, performing cleanup on related logical workers.
        delete_all: Delete all physical worker entities from the database.
        set_db_reference: Set a reference to the main database object for cross-entity operations.
    """

    def create(self, name: str, **kwargs: Any) -> PhysicalWorkerEntity:
        """
        Create a physical worker entity if it does not already exist.

        This method checks whether a physical worker with the specified name exists.
        If not, it creates a new one using the provided attributes.

        Args:
            name (str): The name of the physical worker.
            **kwargs (Any): Additional attributes to pass to the
                [`PhysicalWorkerModel`][db_scripts.data_models.PhysicalWorkerModel] constructor.

        Returns:
            The created or pre-existing physical worker entity.
        """
        log_message_create = f"Physical worker with name '{name}' does not yet have an " "entry in the database. Creating one."
        return self._create_entity_if_not_exists(
            entity_class=PhysicalWorkerEntity,
            model_class=PhysicalWorkerModel,
            identifier=name,
            log_message_exists=f"Physical worker with name '{name}' already has an entry in the database.",
            log_message_create=log_message_create,
            name=name,
            **kwargs,
        )

    def get(self, worker_id_or_name: str) -> PhysicalWorkerEntity:
        """
        Retrieve a physical worker entity by its ID or name.

        Args:
            worker_id_or_name (str): The ID or name of the physical worker to retrieve.

        Returns:
            The physical worker entity corresponding to the provided identifier.

        Raises:
            WorkerNotFoundError: If the specified worker does not exist.
        """
        return self._get_entity(PhysicalWorkerEntity, worker_id_or_name)

    def get_all(self) -> List[PhysicalWorkerEntity]:
        """
        Retrieve all physical worker entities from the database.

        Returns:
            A list of all physical workers stored in the database.
        """
        return self._get_all_entities(PhysicalWorkerEntity, "physical_worker")

    def delete(self, worker_id_or_name: str):
        """
        Delete a physical worker entity by its ID or name.

        This method will also attempt to remove the deleted physical worker's ID
        from its associated logical worker. If the logical worker cannot be found,
        a warning is logged and deletion continues.

        Args:
            worker_id_or_name (str): The ID or name of the physical worker to delete.
        """

        def cleanup_physical_worker(worker):
            logical_worker_id = worker.get_logical_worker_id()
            try:
                logical_worker = self.db.logical_workers.get(worker_id=logical_worker_id)
                logical_worker.remove_physical_worker(worker.get_id())
            except WorkerNotFoundError:
                LOG.warning(
                    f"Couldn't find logical worker with id {logical_worker_id}. Continuing with physical worker delete."
                )

        self._delete_entity(PhysicalWorkerEntity, worker_id_or_name, cleanup_fn=cleanup_physical_worker)

    def delete_all(self):
        """
        Delete all physical worker entities from the database.

        This operation also performs cleanup on associated logical workers as needed.
        """
        self._delete_all_by_type(get_all_fn=self.get_all, delete_fn=self.delete, entity_name="physical workers")

    def set_db_reference(self, db: MerlinDatabase):  # noqa: F821  pylint: disable=undefined-variable
        """
        Set a reference to the main Merlin database object for cross-entity operations.

        This allows the manager to access other entity managers (e.g., runs) when
        performing operations like cleanup during deletions.

        Args:
            db (db_scripts.merlin_db.MerlinDatabase): The database object that provides
                access to related entity managers.
        """
        self.db = db



##############################################################################
# Copyright (c) Lawrence Livermore National Security, LLC and other Merlin
# Project developers. See top-level LICENSE and COPYRIGHT files for dates and
# other details. No copyright assignment is required to contribute to Merlin.
##############################################################################

"""
Module for managing database entities related to physical workers.

This module provides functionality for interacting with physical workers stored in a database,
including creating, retrieving, updating, and deleting them. It defines the `WorkerEntity`
class, which extends the abstract base class [`DatabaseEntity`][db_scripts.entities.db_entity.DatabaseEntity],
to encapsulate worker-specific operations and behaviors.
"""

import logging
from datetime import datetime
from typing import Dict, Optional

from merlin.common.enums import WorkerStatus
from merlin.db_scripts.data_models import PhysicalWorkerModel
from merlin.db_scripts.entities.db_entity import DatabaseEntity
from merlin.db_scripts.entities.mixins.name import NameMixin


LOG = logging.getLogger("merlin")


class PhysicalWorkerEntity(DatabaseEntity[PhysicalWorkerModel], NameMixin):
    """
    A class representing a physical worker in the database.

    This class provides methods to interact with and manage a worker's data, including
    retrieving information about the worker, updating its state, and saving or deleting
    it from the database.

    Attributes:
        entity_info (db_scripts.data_models.PhysicalWorkerModel): An instance of the `PhysicalWorkerModel`
            class containing the physical worker's metadata.
        backend (backends.results_backend.ResultsBackend): An instance of the `ResultsBackend`
            class used to interact with the database.

    Methods:
        __repr__:
            Provide a string representation of the `PhysicalWorkerEntity` instance.
        __str__:
            Provide a human-readable string representation of the `PhysicalWorkerEntity` instance.
        reload_data:
            Reload the latest data for this worker from the database.
        get_id:
            Retrieve the ID of the worker. _Implementation found in
                [`DatabaseEntity.get_id`][db_scripts.entities.db_entity.DatabaseEntity.get_id]._
        get_additional_data:
            Retrieve any additional data saved to this worker. _Implementation found in
                [`DatabaseEntity.get_additional_data`][db_scripts.entities.db_entity.DatabaseEntity.get_additional_data]._
        get_name:
            Retrieve the name of this worker.
        get_logical_worker_id:
            Retrieve the ID of the logical worker that this physical worker was created from.
        get_launch_cmd:
            Retrieve the command used to launch this worker.
        set_launch_cmd:
            Update the launch command used to start this worker.
        get_args:
            Retrieve the arguments for this worker.
        set_args:
            Update the arguments used by this worker.
        get_pid:
            Retrieve the process ID for this worker.
        set_pid:
            Update the process ID for this worker.
        get_status:
            Retrieve the status of this worker.
        set_status:
            Update the status of this worker.
        get_heartbeat_timestamp:
            Retrieve the last heartbeat timestamp of this worker.
        set_heartbeat_timestamp:
            Update the latest heartbeat timestamp of this worker.
        get_latest_start_time:
            Retrieve the time this worker was last started.
        set_latest_start_time:
            Update the latest start time of this worker.
        get_host:
            Retrieve the hostname where this worker is running.
        get_restart_count:
            Retrieve the number of times this worker has been restarted.
        increment_restart_count:
            Increment the restart count for this worker.
        save:
            Save the current state of this worker to the database.
        load:
            (classmethod) Load a `PhysicalWorkerEntity` instance from the database by its ID or name.
        delete:
            (classmethod) Delete a worker from the database by its ID or name.
    """

    @classmethod
    def _get_entity_type(cls) -> str:
        return "physical_worker"

    def __repr__(self) -> str:
        """
        Provide a string representation of the `PhysicalWorkerEntity` instance.

        Returns:
            A human-readable string representation of the `PhysicalWorkerEntity` instance.
        """
        return (
            f"PhysicalWorkerEntity("
            f"id={self.get_id()}, "
            f"name={self.get_name()}, "
            f"logical_worker_id={self.get_logical_worker_id()}, "
            f"launch_cmd={self.get_launch_cmd()}, "
            f"args={self.get_args()}, "
            f"pid={self.get_pid()}, "
            f"status={self.get_status()}, "
            f"heartbeat_timestamp={self.get_heartbeat_timestamp()}, "
            f"latest_start_time={self.get_latest_start_time()}, "
            f"host={self.get_host()}, "
            f"restart_count={self.get_restart_count()}, "
            f"additional_data={self.get_additional_data()}, "
            f"backend={self.backend.get_name()})"
        )

    def __str__(self) -> str:
        """
        Provide a string representation of the `PhysicalWorkerEntity` instance.

        Returns:
            A human-readable string representation of the `PhysicalWorkerEntity` instance.
        """
        worker_id = self.get_id()
        return (
            f"Physical Worker with ID {worker_id}\n"
            f"------------{'-' * len(worker_id)}\n"
            f"Name: {self.get_name()}\n"
            f"Logical Worker ID: {self.get_logical_worker_id()}\n"
            f"Launch Command: {self.get_launch_cmd()}\n"
            f"Args: {self.get_args()}\n"
            f"Process ID: {self.get_pid()}\n"
            f"Status: {self.get_status()}\n"
            f"Last Heartbeat: {self.get_heartbeat_timestamp()}\n"
            f"Last Spinup: {self.get_latest_start_time()}\n"
            f"Host: {self.get_host()}\n"
            f"Restart Count: {self.get_restart_count()}\n"
            f"Additional Data: {self.get_additional_data()}\n\n"
        )

    def get_logical_worker_id(self) -> str:
        """
        Get the ID of the logical worker that this physical worker was created from.

        Returns:
            The ID of the logical worker that this physical worker was created from.
        """
        return self.entity_info.logical_worker_id

    def get_launch_cmd(self) -> str:
        """
        Get the command used to launch this worker.

        Returns:
            The command used to launch this worker.
        """
        return self.entity_info.launch_cmd

    def set_launch_cmd(self, launch_cmd: str):
        """
        Set the launch command used to start this worker.

        Args:
            launch_cmd: The launch command used to start this worker.
        """
        self.entity_info.launch_cmd = launch_cmd
        self.save()

    def get_args(self) -> Dict:
        """
        Get the arguments for this worker.

        Returns:
            A dictionary of arguments for this worker.
        """
        return self.entity_info.args

    def set_args(self, args: str):
        """
        Set the arguments used by this worker.

        Args:
            args: The arguments used by this worker.
        """
        self.entity_info.args = args
        self.save()

    def get_pid(self) -> Optional[int]:
        """
        Get the process ID for this worker.

        Returns:
            The process ID for this worker or None if not set.
        """
        self.reload_data()
        return int(self.entity_info.pid) if self.entity_info.pid else None

    def set_pid(self, pid: str):
        """
        Set the PID of this worker.

        Args:
            pid: The new PID of this worker.
        """
        self.entity_info.pid = pid
        self.save()

    def get_status(self) -> WorkerStatus:
        """
        Get the status of this worker.

        Returns:
            A [`WorkerStatus`][common.enums.WorkerStatus] enum representing
                the status of this worker.
        """
        self.reload_data()
        return self.entity_info.status

    def set_status(self, status: WorkerStatus):
        """
        Set the status of this worker.

        Args:
            status: A [`WorkerStatus`][common.enums.WorkerStatus] enum representing
                the new status of the worker.
        """
        self.entity_info.status = status
        self.save()

    def get_heartbeat_timestamp(self) -> str:
        """
        Get the last heartbeat timestamp of this worker.

        Returns:
            The last heartbeat timestamp we received from this worker
        """
        self.reload_data()
        return self.entity_info.heartbeat_timestamp

    def set_heartbeat_timestamp(self, heartbeat_timestamp: datetime):
        """
        Set the latest heartbeat timestamp of this worker.

        Args:
            heartbeat_timestamp: The latest heartbeat timestamp of this worker.
        """
        self.entity_info.heartbeat_timestamp = heartbeat_timestamp
        self.save()

    def get_latest_start_time(self) -> datetime:
        """
        Get the time that this worker was last started.

        Returns:
            A datetime object representing the last time this worker was started.
        """
        self.reload_data()
        return self.entity_info.latest_start_time

    def set_latest_start_time(self, latest_start_time: datetime):
        """
        Set the latest start time of this worker. This will be set on worker
        startup followed by any time the worker is restarted.

        Args:
            latest_start_time: The latest start time of this worker.
        """
        self.entity_info.latest_start_time = latest_start_time
        self.save()

    def get_host(self) -> str:
        """
        Get the hostname where this worker is running.

        Returns:
            The name of the host that this worker is running on.
        """
        return self.entity_info.host

    def get_restart_count(self) -> int:
        """
        Get the number of times that this worker has been restarted.

        Returns:
            The number of times that this worker has been restarted.
        """
        self.reload_data()
        return self.entity_info.restart_count

    def increment_restart_count(self):
        """
        Add another restart to the restart count.
        """
        self.entity_info.restart_count = self.get_restart_count() + 1
        self.save()
